(* This file is part of Marionnet, a virtual network laboratory
   Copyright (C) 2007  Jean-Vincent Loddo
   Copyright (C) 2007, 2008  Luca Saiu

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. *)


(** Some modules for managing the virtual network *)

open Strings;;
open Sugar;;
open UnixExtra;;
open Environment;;
open Oomarshal;;
open Task_runner;;
open Simple_dialogs;;
open Recursive_mutex;;
open Gettext;;

(** Some constants for drawing with colors. *)
module Color = struct
 let direct_cable    = "#949494" ;;
 let crossover_cable = "#6d8dc0" ;;
end;;

(** A thunk allowing to invoke the sketch refresh method, accessible from many
    modules: *)
let the_refresh_sketch_thunk : (unit -> unit) option ref = ref None;;
let set_refresh_sketch_thunk refresh_sketch_thunk =
  assert(!the_refresh_sketch_thunk = None);
  the_refresh_sketch_thunk := Some refresh_sketch_thunk;;
(** The sketch-refreshing procedure, implemented using the global thunk: *)
let refresh_sketch () =
  match !the_refresh_sketch_thunk with
    Some refresh_sketch_thunk ->
      refresh_sketch_thunk ()
  | None ->
      failwith "get_refresh_sketch_thunk: no global state has been set";;

(* **************************************** *
              Module MSys
 * **************************************** *)

(** Marionnet specific additionnal tools for system operations. *)
module MSys = struct
 type x_policy = HostXServer | Xnest | NoX;;

 let x_policies       = [ HostXServer; Xnest; NoX ];;
 let hostxserver_name = "X HOST";;
 let xnest_name       = "X NEST";;
 let nox_name         = "No X";;

 let x_policy_of_string s =
   if s = hostxserver_name then HostXServer
   else if s = xnest_name then Xnest
   else if s = nox_name then NoX
   else failwith (Printf.sprintf "x_policy_of_string: ill-formed argument %s" s);;

 let string_of_x_policy p =
   match p with
     HostXServer -> hostxserver_name
   | Xnest -> xnest_name
   | NoX   -> nox_name;;

 let x_policies_as_strings = List.map string_of_x_policy x_policies;;

 let marionnet_home             = Initialization.marionnet_home ;;
 let marionnet_home_filesystems = Initialization.marionnet_home_filesystems ;;
 let marionnet_home_kernels     = Initialization.marionnet_home_kernels ;;
 let marionnet_home_images      = Initialization.marionnet_home_images ;;

 let machine_prefix = "machine-" ;;
 let router_prefix  = "router-"  ;;
 let kernel_prefix  = "linux-"   ;;

 let machine_pathname_prefix = marionnet_home_filesystems ^ machine_prefix ;;
 let router_pathname_prefix  = marionnet_home_filesystems ^ router_prefix  ;;
 let kernel_pathname_prefix  = marionnet_home_kernels     ^ kernel_prefix  ;;

 (** Read the given directory searching for names like [prefix ^ "xxxxx"];
     @return the list of ["xxxxx"]. If [suggested_text] is in the list, then
     return it as the {e first} element. If ?extra_element is given then return
     it as the first element (the second if [suggested_text] is present): *)
 let get_unprefixed_file_list ?extra_element prefix directory =
   let prefix_length = String.length prefix in
   let strip_prefix s = String.sub s prefix_length ((String.length s) - prefix_length) in
   let list =
     SysExtra.readdir_into_list
       ~namefilter:(fun file_name -> ((String.length file_name) > prefix_length) &&
                                     ((String.sub file_name 0 prefix_length) = prefix))
       ~nameconverter:strip_prefix
       directory in
   let list = (* Remove directories from the list *)
     List.filter
       (fun file_or_directory ->
         try
           let _ = Sys.readdir (directory ^ "/"^ prefix ^file_or_directory) in
           (* we never get here if readdir raises an exception *)
           false
         with _ ->
           true)
       list in
   let list =
     match extra_element with
       None -> list
     | Some extra_element -> extra_element :: list in
   let list_with_default_first_if_default_is_present =
     if List.exists (fun x -> x = suggested_text) list then
       suggested_text :: (List.filter (fun x -> not (x = suggested_text)) list)
     else
       list in
   list_with_default_first_if_default_is_present;;

 (** Return the list of unprefixed machine filesystem file names: *)
  let machine_filesystem_list () =
    get_unprefixed_file_list
      machine_prefix
      marionnet_home_filesystems;;

 (** Return the list of unprefixed router filesystem file names: *)
  let router_filesystem_list () =
    get_unprefixed_file_list
      router_prefix
      marionnet_home_filesystems;;

 (** Return the list of unprefixed kernel file names: *)
 let kernelList () =
    get_unprefixed_file_list
      kernel_prefix
      marionnet_home_kernels;;

 (** Return the list of unprefixed variant file names for the given (prefixed)
     filesystem; if an entry named like filename_which_should_be_first_if_present,
     when specified, exists, then return it before everything else including
     no_variant_text.*)
 let variant_list_of prefixed_filesystem () =
   get_unprefixed_file_list
     ~extra_element:no_variant_text
     ""
     (marionnet_home_filesystems ^ "/" ^ prefixed_filesystem ^ "_variants");;

 (** Terminal choices to handle uml machines *)
 let termList = List.map string_of_x_policy [HostXServer; Xnest] (*x_policies_as_strings*);;

 (** Kinds for Marionnet related files *)
 type file_kind = Icon | Patch | Network | Distrib | Script ;;

 (** Correct the name of a program resource (not a project resource!)
     according to the install directory of the application *)
 let whereis filename kind = match kind with         (* TO IMPLEMENT *)
 | Icon   -> Initialization.marionnet_home^"/images/"^filename
 |  _     -> filename
 ;;
end;;



(* *************************** *
        Module Dotoptions
 * *************************** *)

module Dotoptions = struct


(* *************************
    class Dotoptions.network
   ************************* *)

type index = int;; (* 0..(length-1) *)
type shuffler = index list ;; (* represents a permutation of indexes of a list*)

(* This part of the state will be filled loading Gui_toolbar_DOT_TUNING. *)
class type dot_tuning_high_level_toolbar_driver =
 object
  method get_iconsize              : string
  method set_iconsize              : string -> unit
  method get_nodesep               : float
  method set_nodesep               : float -> unit
  method get_labeldistance         : float
  method set_labeldistance         : float -> unit
  method get_extrasize             : float
  method set_extrasize             : float -> unit
  method get_image                 : GdkPixbuf.pixbuf
  method get_image_current_width   : int
  method get_image_current_height  : int
  method reset_image_size          : unit -> unit
  method get_image_original_width  : int
  method get_image_original_height : int
end (* class type high_level_toolbar_driver *)

(** Dot options for a network *)
let network_marshaller = new Oomarshal.marshaller;;

class network =

  fun ?(iconsize="large") ?(shuffler=[]) ?(rankdir="TB") ?(nodesep=0.5) ?(labeldistance=1.6) ?(extrasize=0.)

      (* The handler for the real network *)
      (network:( < revertedCables:(string list); revertedCableSet:(bool->string->unit); .. > ))  ->

  object (self)
  inherit Xforest.interpreter ()

  val iconsize = Chip.wref ~name:"iconsize" iconsize
  method iconsize = iconsize

  val rankdir  = Chip.wref ~name:"rankdir" rankdir
  method rankdir = rankdir

  val shuffler = Chip.wref ~name:"shuffler" shuffler
  method shuffler = shuffler

  val nodesep = Chip.wref ~name:"nodesep" nodesep
  method nodesep = nodesep

  val labeldistance = Chip.wref ~name:"labeldistance" labeldistance
  method labeldistance = labeldistance

  val extrasize = Chip.wref ~name:"extrasize" extrasize
  method extrasize = extrasize

  method iconsize_for_dot  = iconsize#get
  method shuffler_as_function = shuffler#get => ListExtra.asFunction  (* returns the permutation function *)
  method rankdir_for_dot   = "rankdir="^(rankdir#get)^";"
  method nodesep_for_dot   = let s=(string_of_float nodesep#get) in ("nodesep="^s^"; ranksep="^s)
  method labeldistance_for_dot = "labeldistance="^(string_of_float labeldistance#get)

  (** This is the method used in user gui callbacks (reactions) *)
  val mutable gui_callbacks_disable : bool   = false
  method gui_callbacks_disable   = gui_callbacks_disable
  method set_gui_callbacks_disable x = gui_callbacks_disable <- x
  method disable_gui_callbacks    () = gui_callbacks_disable <- true
  method enable_gui_callbacks     () =
   ignore (GMain.Timeout.add ~ms:500 ~callback:(fun () -> gui_callbacks_disable <- false; false))

  method reset_shuffler () = shuffler#set []

  method reset_extrasize () =
    begin
    self#toolbar_driver#reset_image_size ();
    extrasize#set 0.;
    ()
    end

  (* Delete _alone here:  *)
  method reset_defaults () =
    begin
      iconsize#set "large";
      shuffler#set [];
      rankdir#set "TB";
      nodesep#set 0.5;
      labeldistance#set 1.6 ;
      ListExtra.foreach network#revertedCables (network#revertedCableSet false) ;
      self#reset_extrasize () ;
      self#set_toolbar_widgets ()
    end

  method ratio : string =
   let extrasize = extrasize#get in
   if (extrasize = 0.) then "ratio=compress;" else
   begin
    let x = self#toolbar_driver#get_image_original_width  => Widget.Image.inch_of_pixels in
    let y = self#toolbar_driver#get_image_original_height => Widget.Image.inch_of_pixels in
    let area  = x *. y in
    let delta_area = extrasize *. area /. 100. in
    let delta = sqrt( (x+.y)**2. +. 4.*. delta_area  )  -.  (x+.y)  in
    let x = x +. delta => string_of_float in
    let y = y +. delta => string_of_float in
    "size=\""^x^","^y^
    "\";\nratio=fill;"
   end

  (** Accessor the dot tuning toolbar. This part of the state will be filled
      loading Gui_toolbar_DOT_TUNING.
      Inverted cables corresponds to dynamic menus, so they not need to be reactualized
      (the dynamic menus are recalculated each time from network#revertedCables. *)

  val mutable toolbar_driver : dot_tuning_high_level_toolbar_driver option = None
  method set_toolbar_driver t = toolbar_driver <- Some t
  method toolbar_driver = match toolbar_driver with Some t -> t | None -> assert false

  (** The dotoption gui reactualization *)

  method set_toolbar_widgets () : unit =
    begin
      self#disable_gui_callbacks   () ;
      self#toolbar_driver#set_iconsize iconsize#get ;
      self#toolbar_driver#set_nodesep nodesep#get ;
      self#toolbar_driver#set_labeldistance labeldistance#get ;
      self#toolbar_driver#set_extrasize extrasize#get ;
      self#enable_gui_callbacks    () ;
      ()
    end

  (** Marshalling is performed in this ugly way because directly dumping the whole [self] object
      would involve resolving references to Gtk callbacks, which are outside the OCaml heap and
      hence (understandably) not supported by the marshaller. *)

  (** Dump the current state of [self] into the given file. *)
  method save_to_file (file_name : string) =
    Xforest.print_forest network#to_forest;
    network_marshaller#to_file self#to_forest file_name

  (** This method is used just for undumping dotoptions, so is not strict.
      For instance, exceptions provoked by bad cable names are simply ignored. *)
  method set_revertedCables names =
    ListExtra.foreach names (fun n -> try (network#revertedCableSet true n) with _ -> ())

  (** Undump the state of [self] from the given file. *)
  method load_from_file (file_name : string) =
   let forest = network_marshaller#from_file file_name in
   Xforest.print_forest forest;
   match forest with
   | Forest.NonEmpty  (("dotoptions", attrs) , childs , Forest.Empty) ->
      self#from_forest ("dotoptions", attrs) childs

 (** Dotoptions to forest encoding. *)
  method to_forest =
   Forest.leaf ("dotoptions", [
    		   ("iconsize"      , iconsize#get                   ) ;
    		   ("shuffler"      , (Xforest.encode shuffler#get)  ) ;
                   ("rankdir"       , rankdir#get                    ) ;
                   ("nodesep"       , (string_of_float nodesep#get)      ) ;
                   ("labeldistance" , (string_of_float labeldistance#get)) ;
                   ("extrasize"     , (string_of_float extrasize#get)    ) ;
                   ("gui_callbacks_disable", (string_of_bool gui_callbacks_disable)) ;
                   ("invertedCables", (Xforest.encode network#revertedCables)) ;
	           ])

 (** A Dotoption.network has just attributes (no childs) in this version.
     The Dotoption.network must be undumped AFTER the Netmodel.network in
     order to have significant cable names (revertedCables). *)
 method eval_forest_attribute = function
  | ("iconsize"             , x ) -> self#iconsize#set       x
  | ("shuffler"             , x ) -> self#shuffler#set      (Xforest.decode x)
  | ("rankdir"              , x ) -> self#rankdir#set        x
  | ("nodesep"              , x ) -> self#nodesep#set       (float_of_string x)
  | ("labeldistance"        , x ) -> self#labeldistance#set (float_of_string x)
  | ("extrasize"            , x ) -> self#extrasize#set     (float_of_string x)
  | ("gui_callbacks_disable", x ) -> self#set_gui_callbacks_disable (bool_of_string x)
  | ("invertedCables"       , x ) -> self#set_revertedCables (Xforest.decode x)
  | _ -> () (* Forward-comp. *)

end;; (* class Dotoptions.network *)


(** Dot options for a cable *)
class cable ?(reverted=false) ~(motherboard:State_types.motherboard) ~name () =
 object (self)
(*   val reverted = motherboard#reverted_rj45cables_cable#create_wref ~name:(name^"_reverted") reverted*)
   val reverted = Chip.wref_in_cable ~name:(name^"_reverted") ~cable:motherboard#reverted_rj45cables_cable reverted
   method reverted = reverted
   method destroy =
    reverted#destroy
 end;; (* class Dotoptions.cable *)


end;; (* module Dotoptions *)


(* **************************************** *
              Module Netmodel
 * **************************************** *)


(** Model for managing the virtual network defined step by step by user *)
module Netmodel = struct

(** {2 Basic functions and types } *)

(** A device may be a Hub, a Switch or a Router. *)
type devkind   = Hub | Switch | Router | NotADevice ;;

(** A cable may be Direct, Crossover or NullModem. *)
type cablekind = Direct | Crossover | NullModem ;;

(** A port (of machine or device) may Eth (ethernet) or Ttys (serial). *)
type portkind  = Eth | TtyS ;;

(** A node is a Machine, a Device, a Cloud or a GwInternet.
    Cable are edge in the network, which is a graph of these nodes. *)
type nodekind  = Machine | Device | Cloud | Gateway ;;

(** String conversion for a devkind. *)
let string_of_devkind = function
  | Hub        -> "hub"
  | Switch     -> "switch"
  | Router     -> "router"
  | NotADevice -> raise (Failure "string_of_devkind: NotADevice")
;;

(** String conversion for a cablekind. *)
let string_of_cablekind = function
  | Direct     -> "direct"
  | Crossover  -> "crossover"
  | NullModem  -> "nullmodem"
;;

(** String conversion for a portkind. *)
let string_of_portkind = function
  | Eth   -> "eth"
;;

(** The portkind interpretation of the given string. *)
let devkind_of_string x = match x with
  | "hub"     -> Hub
  | "switch"  -> Switch
  | "router"  -> Router
  | _         -> raise (Failure ("devkind_of_string"^x))
;;

(** The cablekind interpretation of the given string. *)
let cablekind_of_string x = match x with
  | "direct"    -> Direct
  | "crossover" -> Crossover
  | "nullmodem" -> NullModem
  | "serial"    -> NullModem
  | _           -> raise (Failure ("cablekind_of_string: "^x))
;;


(** Examples: pc1, pc2 if the node is a machines, A,B,C if the node is a device *)
type nodename   = string ;;

(** Examples: eth0..eth4 *)
type receptname = string ;;

(** A socketname is a pair (nodename,receptname), for instance ("rome";"eth0")
    or ("H1";"eth2"). *)
type socketname = { mutable nodename:string ; mutable receptname:string ; } ;;

type name   = string ;;
type label  = string ;;
let nolabel = "";;

(** iconsize may be "small", "med", "large" or "xxl". *)
type iconsize = string ;;

(** {2 Classes} *)

type simulated_device_automaton_state =
   NoDevice         (** *)
 | DeviceOff        (** *)
 | DeviceOn         (** *)
 | DeviceSleeping   (** *)
;;

exception ForbiddenTransition;;

(** This represents the current state of a simulated device (as per
    network_simulation.ml) and enables easy high-level state transitions which
    conveniently hide the complexity of managing switches and cables; when the
    user tries to invoke any forbidden state transition an exception is
    raised. *)
class virtual simulated_device = object(self)
  (** We have critical sections here: *)
  val mutex = Recursive_mutex.create ()

  (** The current automaton state, and its access method: *)
  val automaton_state = ref NoDevice

  (** Get the state of simulated device. *)
  method simulated_device_state =
    !automaton_state

  (** This string will be used to select the good icon for the dot sketch. *)
  method string_of_simulated_device_state = match !automaton_state with
  | DeviceOff      -> "off"
  | DeviceOn       -> "on"
  | DeviceSleeping -> "pause"
  | _              -> "off" (* Sometimes the sketch is builded in this state, so... *)

  (** The automaton state this device is going to. This is only used for the GUI and
      is not guaranteed to be accurate in case of concurrent access. It's only
      guaranteed to always hold some value of the correct type.
      If no transition is occurring then the ref should hold None. *)
  val next_automaton_state = ref (Some NoDevice)

  method next_simulated_device_state =
    !next_automaton_state

  method set_next_simulated_device_state state =
    next_automaton_state := state;
    refresh_sketch (); (* show our transient simulation state icon *)

  method virtual get_name : string

  (** The device implementing the object in the simulated network, if any (this is
      ref None when the device has not been started yet, or some state modification
      happened) *)
  val simulated_device : Simulated_network.device option ref =
    ref None

  (** Create a new simulated device according to the current status *)
  method virtual make_simulated_device : Simulated_network.device

  (** Return the list of cables directly linked to a port of self as an endpoint.
      This is needed so that simulated cables can be automatically started/destroyed
      as soon as both their endpoints are created/destroyed *)
  method private get_involved_cables = []

  (** Return a hublet process by name, so that we're able to easily find where to
      connect/disconnect cables. This fails if the device is non-existing, as there
      are no hublets in this case *)
  method get_hublet_process_by_interface_name name =
    match !simulated_device with
      Some(sd) -> sd#get_ethernet_port_by_name name
    | None -> failwith "looking for a hublet when its device is non-existing"

  (** Return true iff hublet processes are currently existing. This is only meaningful
      for devices which can actually have hublets *)
  method has_hublet_processes =
    match !simulated_device with
      Some(_) -> true
    | None -> false

  method private enqueue_task_with_progress_bar verb thunk =
    let text = verb ^ " " ^ self#get_name in
    let progress_bar = ref None in
    the_task_runner#schedule
      ~name:text
      (fun () ->
        (try
          progress_bar := Some (make_progress_bar_dialog ~title:text ());
          thunk ();
        with e -> begin
          Log.printf "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
          let message =
            Printf.sprintf "enqueue_task_with_progress_bar: %s %s failed (%s)"
              verb self#get_name (Printexc.to_string e) in
          Log.printf "%s\n" message;
          Simple_dialogs.warning message message ();
          Log.printf "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
          flush_all ();
        end));
    the_task_runner#schedule
      ~name:("Destroy the progress bar for \"" ^ text ^ "\"")
      (fun () ->
        match !progress_bar with
          Some progress_bar ->
            destroy_progress_bar_dialog progress_bar
        | None ->
            assert false)

(*  method reboot_if_possible =
    try
      self#suspend;
      self#startup;
    with _ ->
      () *)

  method create =
    (* This is invisible for the user: don't set the next state *)
    the_task_runner#schedule ~name:("create "^self#get_name) (fun () -> self#create_right_now)

  method destroy =
    (* This is invisible for the user: don't set the next state *)
    the_task_runner#schedule ~name:("destroy "^self#get_name)(fun () -> self#destroy_right_now)

  method startup =
    self#set_next_simulated_device_state (Some DeviceOn);
    self#enqueue_task_with_progress_bar (s_ "Starting") (fun () -> if self#can_startup then self#startup_right_now)

  method suspend =
    self#set_next_simulated_device_state (Some DeviceSleeping);
    self#enqueue_task_with_progress_bar (s_ "Suspending") (fun () -> if self#can_suspend then self#suspend_right_now)

  method resume =
    self#set_next_simulated_device_state (Some DeviceOn);
    self#enqueue_task_with_progress_bar (s_ "Resuming") (fun () -> if self#can_resume then self#resume_right_now)

  method gracefully_shutdown =
    self#set_next_simulated_device_state (Some DeviceOff);
    self#enqueue_task_with_progress_bar (s_ "Stopping") (fun () -> if self#can_gracefully_shutdown then self#gracefully_shutdown_right_now)

  method poweroff =
    self#set_next_simulated_device_state (Some DeviceOff);
    self#enqueue_task_with_progress_bar (s_ "Shutting down") (fun () -> if self#can_poweroff then self#poweroff_right_now)

  method (*private*) create_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("+ About to create the simulated device " ^ (self#get_name) ^ "\n");
        Log.print_string ("  (it's connected to " ^
                      (string_of_int (List.length (self#get_involved_cables))) ^
                      " cables)\n");
        match !automaton_state, !simulated_device with
          NoDevice, None -> (simulated_device := (Some self#make_simulated_device);
                             automaton_state := DeviceOff;
                             self#set_next_simulated_device_state None;
                             (* An endpoint for cables linked to self was just added; we
                                may need to start some cables. *)
                             ignore (List.map
                                       (fun cable -> Log.print_string ("+ Working on cable\n"^(cable#show "")^"\n");
                                         cable#increment_alive_endpoints_no)
                                       (self#get_involved_cables)))
        | _ -> raise ForbiddenTransition)

  (** The unit parameter is needed: see how it's used in simulated_network: *)
  method private destroy_because_of_unexpected_death () =
    Log.printf "You don't deadlock here %s, do you? -1\n" self#get_name; flush_all ();
    with_mutex mutex
      (fun () ->
        Log.printf "You don't deadlock here %s, do you? 0\n" self#get_name;
        flush_all ();
        (try
          self#destroy_right_now
        with e -> begin
          Log.printf "WARNING: destroy_because_of_unexpected_death: failed (%s)\n"
            (Printexc.to_string e);
        end;
          self#set_next_simulated_device_state None)); (* don't show next-state icons for this *)

  method (*private*) destroy_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("- About to destroy the simulated device " ^ (self#get_name) ^ "\n");
        match !automaton_state, !simulated_device with
          (DeviceOn | DeviceSleeping), Some(d) ->
            (Log.print_string ("  (destroying the on/sleeping device " ^ (self#get_name) ^ ". Powering it off first...)\n");
             self#poweroff_right_now; (* non-gracefully *)
             self#destroy_right_now)
        | NoDevice, None ->
            Log.print_string ("  (destroying the already 'no-device' device " ^ (self#get_name) ^ ". Doing nothing...)\n");
            () (* Do nothing, but don't fail. *)
        | DeviceOff, Some(d) ->
            ((* An endpoint for cables linked to self was just added; we
                may need to start some cables. *)
             Log.print_string ("  (destroying the off device " ^ (self#get_name) ^ ": decrementing its cables rc...)\n");
             List.iter
               (fun cable ->
                 Log.print_string ("- Unpinning the cable "^(cable#show "")^"\n"); flush_all ();
                 cable#decrement_alive_endpoints_no;
                 Log.print_string ("- The cable "^(cable#show "")^" was unpinned with success\n"); flush_all ())
               self#get_involved_cables;
             Log.print_string ("  (destroying the simulated device implementing " ^ (self#get_name) ^ "...)\n");
             d#destroy; (* This is the a method from some object in Simulated_network *)
             simulated_device := None;
             automaton_state := NoDevice;
             self#set_next_simulated_device_state None;
             Log.printf "- We're not deadlocked yet (%s). Great.\n" self#get_name; flush_all ());
        | _ ->
            raise ForbiddenTransition);
    Log.print_string ("- The simulated device " ^ (self#get_name) ^ " was destroyed with success\n");
    flush_all ()


  method (*private*) startup_right_now =
    with_mutex mutex
      (fun () ->
        (* Don't startup ``incorrect'' devices. This is currently limited to cables of the
           wrong crossoverness which the user has defined by mistake: *)
        if self#is_correct then begin
          Log.print_string ("* starting up the device " ^ (self#get_name) ^ "...\n");
          match !automaton_state, !simulated_device with
            NoDevice, None -> (Log.print_string ("  (creating processes for " ^ (self#get_name) ^ " first...)\n");
                               self#create_right_now;
                               Log.print_string ("  (processes for " ^ (self#get_name) ^ " were created...)\n");
                               self#startup_right_now)
          | DeviceOff, Some(d) -> (d#startup;  (* This is the a method from some object in Simulated_network *)
                                   automaton_state := DeviceOn;
                                   self#set_next_simulated_device_state None;
                                   Log.print_string ("* The device " ^ (self#get_name) ^ " was started up\n"))
          | _ -> raise ForbiddenTransition
        end else begin
          Log.print_string ("* REFUSING TO START UP the ``incorrect'' device " ^ (self#get_name) ^ "...\n");
        end)

  method (*private*) suspend_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("|| Suspending up the device " ^ (self#get_name) ^ "...\n");
        match !automaton_state, !simulated_device with
          DeviceOn, Some(d) -> (d#suspend; (* This is the a method from some object in Simulated_network *)
                                automaton_state := DeviceSleeping;
                                self#set_next_simulated_device_state None)
        | _ -> raise ForbiddenTransition)

  method (*private*) resume_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("|> Resuming the device " ^ (self#get_name) ^ "...\n");
        match !automaton_state, !simulated_device with
          DeviceSleeping, Some(d) -> (d#resume; (* This is the a method from some object in Simulated_network *)
                                      automaton_state := DeviceOn;
                                      self#set_next_simulated_device_state None)
        | _ -> raise ForbiddenTransition)

  method (*private*) gracefully_shutdown_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("* Gracefully shutting down the device " ^ (self#get_name) ^ "...\n");
        match !automaton_state, !simulated_device with
          DeviceOn, Some(d) -> (d#gracefully_shutdown; (* This is the a method from some object in Simulated_network *)
                                automaton_state := DeviceOff;
                                self#set_next_simulated_device_state None)
        | DeviceSleeping, Some(d) -> (self#resume_right_now;
                                      self#gracefully_shutdown_right_now)
        | _ -> raise ForbiddenTransition)

  method (*private*) poweroff_right_now =
    with_mutex mutex
      (fun () ->
        Log.print_string ("* Powering off the device " ^ (self#get_name) ^ "...\n");
        match !automaton_state, !simulated_device with
          DeviceOn, Some(d) -> (d#shutdown; (* non-gracefully *)
                                automaton_state := DeviceOff;
                                self#set_next_simulated_device_state None)
        | DeviceSleeping, Some(d) -> (self#resume_right_now;
                                      self#poweroff_right_now)
        | _ -> raise ForbiddenTransition)

  method set_hublet_processes_no n =
    with_mutex mutex
      (fun () ->
        Log.print_string ("* Updating the number of hublets of the device " ^ (self#get_name) ^ "...\n");
        match !automaton_state, !simulated_device with
          DeviceOff, Some(d) -> d#set_hublet_processes_no n (* update hublets and don't change state *)
        | NoDevice, None -> (self#create_right_now;                    (* Make hublets... *)
                             self#set_hublet_processes_no n) (* ...and do the real work in state Off *)
        | _ ->  raise ForbiddenTransition)

  (** Return true iff the current state allows to 'startup' the device from the GUI. *)
  method can_startup =
    with_mutex mutex
      (fun () ->
        match !automaton_state with NoDevice | DeviceOff -> true | _ -> false)

  (** Return true iff the current state allows to 'shutdown' a device from the GUI. *)
  method can_gracefully_shutdown =
    with_mutex mutex
      (fun () ->
        match !automaton_state with DeviceOn | DeviceSleeping -> true | _ -> false)

  (** Return true iff the current state allows to 'power off' a device from the GUI. *)
  method can_poweroff =
    with_mutex mutex
      (fun () ->
        match !automaton_state with NoDevice | DeviceOff -> false | _ -> true)

  (** Return true iff the current state allows to 'suspend' a device from the GUI. *)
  method can_suspend =
    with_mutex mutex
      (fun () ->
        match !automaton_state with DeviceOn -> true | _ -> false)

  (** Return true iff the current state allows to 'resume' a device from the GUI. *)
  method can_resume =
    with_mutex mutex
      (fun () ->
        match !automaton_state with DeviceSleeping -> true | _ -> false)

  (** 'Correctness' support: this is needed so that we can refuse to start incorrectly
      placed components such as Ethernet cables of the wrong crossoverness, which the user
      may have created by mistake: *)
  val is_correct =
    ref true (* devices are ``correct'' by default *)
  method is_correct =
    with_mutex mutex
      (fun () ->
        !is_correct)
  method set_correctness correctness =
    with_mutex mutex
      (fun () ->
        is_correct := correctness)
end;;


(* *************************** *
      class common
 * *************************** *)

(** The shared generator for all unique ids: *)
let generator = Counter.make_int_generator ();;

(** General-purpose class with common fields as
    - an automatically generated identifier
    - a (mutable) name; i.e. valid string identifier
    - a (mutable) label (string) not containing '<' and '>' (for prevent conflict with dot)
*)
class id_name_label = fun ?(name="noname") ?(label="") () ->

  (* Some checks over used name and label *)
  let wellFormedLabel =  (StrExtra.Bool.match_string ".*[><].*") || not in

  let check_name  x =
  	if not (StrExtra.wellFormedName  x)
       	then failwith ("Setting component "^name^": invalid name")
        else x in

  let check_label x =
  	if not (wellFormedLabel x)
        then failwith ("Setting component "^name^": invalid label")
        else (StringExtra.strip x) in

  object (self)

  (** A component has an immutable identifier. *)
  val id : int = generator ()
  method id = id

  (** A component has a mutable name. *)
  val mutable name : string = check_name name

  method get_name   = name
  method set_name x = name <- check_name x

  (* A shortcut for get_name *)
  method      name  = name

  (** A component has a label, usually used by dot drawing the network graph. *)
  val mutable label : string = check_label label

  method get_label : string = label
  method set_label x = label <- check_label x

end;;

(* *************************** *
        class component
 * *************************** *)

(** A component may be a node (machine or device) or a cable (direct, crossover or nullmodem).
    It's simply a thing with a name and an associated (mutable) label. *)
class virtual component =
fun ~(network:< .. >)
    ?(name="noname")
    ?(label="") () ->

  object (self)
  inherit id_name_label ~name ~label ()
  inherit Xforest.interpreter ()

  (** The global network. It's a pain to have to access this via a global variable
      define in Marionnet *)
  val network = network

end;;


(* *************************** *
          class receptacle
 * *************************** *)

(** A receptacle is the abstract notion representing both ethernet and serial ports.
    It's a component with a portkind [(Eth|TtyS)]. *)
class receptacle = fun ~network (name:receptname) (index:int) label (kind:portkind) ->

  (* Some checks over receptacle name *)

   let wellFormedReceptName k x =
    match k with
    | Eth  -> (StrExtra.Bool.match_string "^eth[0-9]$"        x) or
              (StrExtra.Bool.match_string "^eth[1-9][0-9]+$"  x) or
              (StrExtra.Bool.match_string "^port[0-9]$"       x) or
              (StrExtra.Bool.match_string "^port[1-9][0-9]+$" x) in

  let check_receptname k x = if not (wellFormedReceptName k x) then
    failwith ("Setting receptacle: name "^x^" not valid for this kind of receptacle")
    else x in

  object (self)

  inherit id_name_label ~name:(check_receptname kind name) ~label ()

  (** The kind of receptacle (Eth|TtyS) *)
  method portkind = kind

  (** A receptacle has an index in its container. This field is normally coherent with the label.
      For instance, one expects that a receptacle with the label "eth2" will have the index=2. *)
  method index = if (index>=0)
                 then index
                 else name => ((StrExtra.extract_groups (Str.regexp "[a-z]*\\([0-9]+\\)")) || List.hd || int_of_string)
end;;


(* *************************** *
          class node
 * *************************** *)

(** Machines and routers have MDI ports, switches and hubs have MDI_X a priori.
    Currently, devices are sold with "intelligent" ports, i.e. MDI/MDI-X. *)
type polarity = MDI | MDI_X | Intelligent ;;

(** A node of the network is a container of receptacles.
    Defects may be added after the creation, using the related method. *)
class virtual node = fun
   ~network
   ?(name="nonodename")
   ?(label="")
   ?(nodekind = Machine)
   ?(devkind  = NotADevice)
   ?(rlist:receptacle list = []) () ->

   object (self)
   inherit component ~network ~name ~label ()
   inherit simulated_device

  (** 'Static' methods (in the sense of C++/Java). Polarity is used to decide the correct
      kind of Ethernet cable needed to connect a pair of devices: the cable should be
      crossover iff both endpoints have the same polarity: *)
  method virtual polarity : polarity

  (** Constant fields (methods) *)

  (** The mutable list of receptacle of this component. *)
  val mutable nodekind = nodekind
  val mutable devkind  = devkind

  method set_nodekind x = nodekind <- x
  method set_devkind  x = devkind  <- x

  (** The kind of the node (Machine or Device). Alias for get_kind.*)
  method nodekind = nodekind

  (** The kind of the device (if the node is a device). *)
  method devkind = devkind

  (** The method which give access to the object describing defects of this component. *)
  (* method defect = defect *)

  (** Variable fields *)

  (** The mutable list of receptacle of this component. *)
  val mutable receptacles : (receptacle list) = rlist

  method set_receptacles x = receptacles <- x

  (** Retrieving informations about receptacles *)

  (** Basic accessor for the list of receptacles of this node. The portkind may be specified. *)
  method get_receptacles ?(portkind:(portkind option)=None) () =
    match portkind with
    | None   -> receptacles
    | Some k -> List.filter (fun r->r#portkind=k) receptacles

  (** Get a receptacle of this node by the name of this receptacle (for instance ["eth0"]). *)
  method getReceptacleByName rname = List.find (fun x->x#name=rname) receptacles

  (** A receptacle of the given name exists in this node? *)
  method receptacleExists n = let f=(fun x->x#name=n) in (List.exists f receptacles)

  (** The number of receptacles of this node. The portkind may be specified. *)
  method numberOfReceptacles ?(portkind:(portkind option)=None) ()
      = List.length (self#get_receptacles ~portkind ())

  (** The max index in the list of receptacles of this node. The portkind is here mandatory. *)
  method maxReceptacleIndex (k:portkind)  =
    let l = (List.map (fun x->x#index) (self#get_receptacles ~portkind:(Some k) ())) in
    if l=[] then -1 else (ListExtra.max l)

  (** The list of names of receptacles of this node. The portkind may be specified. *)
  method receptacleNames ?(portkind:(portkind option)=None) () =
    List.map (fun x->x#name) (self#get_receptacles ~portkind ())

  (** The list of socketnames [\{nodename:string; receptname:string\}]
      contained in this node. The [nodename] field will be constant to the name of this node in the returned list.
      The portkind may be specified. *)
(*  method socketnames ?(portkind:(portkind option)=None) () =
    List.map (fun x->{nodename=self#name; receptname=x}) (self#receptacleNames ~portkind ()) *)
  method socketnames =
    List.map (fun x->{nodename=self#name; receptname=x}) (self#receptacleNames ())

  (** Just a version of socketnames returning something easier to manage than a list *)
  method socketnames_as_pair =
    match self#socketnames with
      [ a; b ] ->
        a, b
    | _ ->
        assert false

  (** Adding and removing receptacles *)

  (** Append a receptacle at the end of the receptacles list of this node. This action is performed
      checking the unicity of the receptacle name. *)
  method append_receptacle r  =
    let name_exists = List.exists (fun x->x=r#name) (self#receptacleNames ()) in
    if  name_exists
          then raise (Failure "add_receptacle : name already exists in this node")
          else receptacles <- (receptacles@[r])

  (** Append a receptacle at the end of the receptacles list of this node. This action is performed
      using a fresh index and name for the created receptacle. The portkind is here mandatory. *)
  method add_progressive_receptacle ?(prefix=None) (k:portkind) =
    let num = (self#maxReceptacleIndex k) + 1 in
    let pref = (match prefix with None -> (string_of_portkind k) | Some x -> x) in
    let name = pref^(string_of_int num) in
    let r = new receptacle network name num "" k in
    self#append_receptacle r


  (** Remove the last receptacle (with the maximal index) of the given portkind from this node. *)
  method rm_last_receptacle (k:portkind) =
    let i = self#maxReceptacleIndex k in
    let concerned = (fun r->(r#portkind=k) && (r#index=i)) in
    receptacles <- List.filter (fun r -> not (concerned r)) receptacles

  (** Re-adjusting the number of receptacles *)

  (** Set the number of receptacles of the given [portkind] to the given number [x].
      If the current number of receptacles is greater than [x] then append else remove the necessary number
      of receptacles.*)
  method readjust_receptacleNumber ?(prefix=None) (k:portkind) x =
    let n = (self#numberOfReceptacles ~portkind:(Some k) ()) in
    match (n>x,n<x) with
    | (true,  _ )  -> (self#rm_last_receptacle k)                 ; self#readjust_receptacleNumber ~prefix k x
    | ( _  ,true)  -> (self#add_progressive_receptacle ~prefix k) ; self#readjust_receptacleNumber ~prefix k x
    |      _       -> ()


  (** {b Convenient aliases} *)

  method get_eth_number                   = self#numberOfReceptacles ~portkind:(Some Eth) ()
  method set_eth_number ?(prefix="eth") x = self#readjust_receptacleNumber ~prefix:(Some prefix) Eth x

  (** Node dot traduction *)

  (** Returns an image representig the node with the given iconsize. *)
  method virtual dotImg : iconsize -> string

  (** Returns the label to use for cable representation.
      This method may be redefined (for instance in [gateway]). *)
  method dotLabelForEdges (receptname:string) = self#get_label

  (** Returns the port to use for cable representation.
      This method may be redefined (for instance in [gateway]). *)
  method dotPortForEdges (receptname:string)  = receptname

  (** A node is represented in dot with an HTML label which is a table
      with a first line containing the name, with a second line containing the node associated image (method [dotImg]),
      and, if the node has a label, a third line containing the label. With the [nodeoptions] parameter one can force,
      for example, the fontsize or fontname for both name and label :
      [ dotTrad ~nodeoptions="fontsize=8" "large" ] *)
  method dotTrad ?(nodeoptions="") (z:iconsize) =
    let label_line = if self#get_label="" then "" else "<TR><TD><FONT COLOR=\"#3a3936\">"^self#get_label^"</FONT></TD></TR>" in
    let fontsize   = if self#nodekind=Machine then "" else "fontsize=8," in
    let nodeoptions = if nodeoptions = "" then "" else (nodeoptions^",") in
    begin
    self#name^" ["^fontsize^nodeoptions^"shape=plaintext,label=<
<TABLE BORDER=\"0\" CELLBORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"0\">
  <TR><TD>"^self#name^"</TD></TR>
  <TR><TD PORT=\"img\"><IMG SRC=\""^(self#dotImg z)^"\"></IMG></TD></TR>
"^label_line^"
</TABLE>>];"
    end

  (** make_simulated_device is defined in subclasses, not here  *)

  (* Return the list of cables of which a port of self is an endpoint: *)
  method private get_involved_cables =
    List.filter (fun c->c#isNodeinvolved self#get_name) network#cables

  (** I need these even if such mechods are only useful for machines (and routers,
      in the future), because of type problems... This looks like the easiest way. *)
  method get_distrib : string =
    failwith ("The device " ^ self#get_name ^ " has not a type supporting distribution")

  method get_variant : string =
    failwith ("The device " ^ self#get_name ^ " has not a type supporting variants")

  method set_variant : string -> unit =
    failwith ("The device " ^ self#get_name ^ " has not a type supporting variants")
  (** Don't store the variant name as a symlink; resolve it if it's a symlink: *)

  method resolve_variant =
    try (* Let's support this on *every* device; it's simpler... *)
      let variant = self#get_variant in
      (try
        let marionnet_home_filesystems = Initialization.marionnet_home_filesystems in
        let prefixed_filesystem = "machine-" ^ self#get_distrib in
        let variant_pathname =
          marionnet_home_filesystems ^ "/" ^ prefixed_filesystem ^ "_variants/" ^ variant in
        let resolved_variant =
          Filename.basename (Unix.readlink variant_pathname) in
        self#set_variant resolved_variant;
        Log.printf "The variant \"%s\" was a symlink. Resolved into \"%s\".\n" variant resolved_variant;
        flush_all ();
      with _ ->
        Log.printf "The variant \"%s\" is not a symlink.\n" variant; flush_all ());
    with _ ->
      ();
end;;


(* *************************** *
        class cable
 * *************************** *)

(** A cable defines an edge in the network graph. It is defined by two socketnames (mutables), i.e. two records
   in the form [ { nodename:string; receptname:string } ] representing the connection provided by the cable.
   Defects may be added after creation. *)
class cable =
   fun
   ~network
   ~motherboard
   ?(name="nocablename")
   ?(label="")
   ~cablekind
   ~(left :socketname)
   ~(right:socketname)
   () ->
  let dotoptions = new Dotoptions.cable ~motherboard ~name () in
  let with_mutex mutex thunk =
    try
      let result = thunk () in
      result
    with e -> begin
      Log.printf "I disabled synchronization here: RE-RAISING %s\n" (Printexc.to_string e); flush_all ();
      raise e;
    end
  in
object (self)
  inherit component ~network ~name ~label ()
  inherit simulated_device as super_simulated_device

  (* Redefinition: *)
  method destroy =
    super_simulated_device#destroy;
    self#dotoptions#destroy ()

  val cablekind = cablekind
  method cablekind = cablekind

  method dotoptions = dotoptions

  (** A cable may be either connected or disconnected; it's connected by default: *)
  val connected =
    ref true

  (** A cable has two connected socketnames. *)
  val mutable left  : socketname = left
  val mutable right : socketname = right

  (** Accessors *)
  method get_left    =
    with_mutex mutex
      (fun () -> left)
  method get_right   =
    with_mutex mutex
      (fun () -> right)
  method set_left  x =
    with_mutex mutex
      (fun () -> left  <- x)
  method set_right x =
    with_mutex mutex
      (fun () -> right <- x)

  (** The list of concerned socketnames *)
  method socketnames =
    with_mutex mutex
      (fun () ->
        [self#get_left; self#get_right])

  (** Just a version of socketnames returning something easier to manage than a list *)
  method socketnames_as_pair =
    with_mutex mutex
      (fun () ->
        match self#socketnames with
          [ a; b ] ->
            a, b
        | _ ->
            assert false)

  (** The li st of two names of nodes (machine/device) linked by the cable *)
  method involvedNodeNames =
    with_mutex mutex
      (fun () ->
        [self#get_left.nodename; self#get_right.nodename])

  (** The pair of nodes (machine/device) linked by the cable *)
  method involvedNodes =
    with_mutex mutex
      (fun () ->
        (network#getNodeByName self#get_left.nodename),
        (network#getNodeByName self#get_right.nodename))

  (** Return the list of devices (i.e. hubs, switches or routers) directly linked to this cable: *)
  method involved_devices_and_port_nos =
    with_mutex mutex
      (fun () ->
        let left_port_name, right_port_name = self#receptacle_names in
        let left_port_index = (self#left_endpoint#getReceptacleByName left_port_name)#index in
        let right_port_index = (self#right_endpoint#getReceptacleByName right_port_name)#index in
        let left_node =
          network#getNodeByName self#get_left.nodename in
        let right_node =
          network#getNodeByName self#get_right.nodename in
        let left_device =
          match left_node#devkind with
            NotADevice -> None
          | _ -> Some (left_node, left_port_index) in
        let right_device =
          match right_node#devkind with
            NotADevice -> None
          | _ -> Some (right_node, right_port_index) in
        match left_device, right_device with
          None, None ->     []
        | None, (Some d) -> [ d ]
        | (Some d), None -> [ d ]
        | (Some d1), (Some d2) -> [ d1; d2 ])

  (** Is a node connected to something with this cable? *)
  method isNodeinvolved nodename =
    with_mutex mutex
      (fun () ->
        List.mem nodename self#involvedNodeNames)

  (** Change the left or right nodename if needed. *)
  method changeNodeName oldname newname =
    with_mutex mutex
      (fun () ->
        if left.nodename  = oldname then left.nodename  <- newname ;
        if right.nodename = oldname then right.nodename <- newname)

  (** Look for socktname in the left or right side of the cable and return the other socketname if exists *)
  method partnerOf (x:socketname) =
    with_mutex mutex
      (fun () ->
        if (x=self#get_left) then self#get_right
        else
          (if (x=self#get_right) then self#get_left else raise (Failure "partnerOf")))

  (** Show its definition. Useful for debugging. *)
  method show prefix =
    with_mutex mutex
      (fun () ->
        prefix^self#name^" ("^(string_of_cablekind self#cablekind)^")"^
        " ["^self#get_left.nodename^","^self#get_left.receptname^"] -> "^
        " ["^self#get_right.nodename^","^self#get_right.receptname^"]")

  method to_forest =
    with_mutex mutex
      (fun () ->
        Forest.leaf ("cable",
                       [ ("name"            ,  self#get_name )  ;
   		         ("label"           ,  self#get_label)  ;
                         ("kind"            , (string_of_cablekind self#cablekind)) ;
                         ("leftnodename"    ,  self#get_left.nodename)    ;
                         ("leftreceptname"  ,  self#get_left.receptname)  ;
                         ("rightnodename"   ,  self#get_right.nodename)   ;
                         ("rightreceptname" ,  self#get_right.receptname) ;
                       ]))

  (** A cable has just attributes (no childs) in this version. The attribute "kind" cannot be set,
      must be considered as a constant field of the class. *)
  method eval_forest_attribute =
    function
      | ("name"            , x ) -> self#set_name  x
      | ("label"           , x ) -> self#set_label x
      | ("leftnodename"    , x ) -> self#get_left.nodename    <- x ;
      | ("leftreceptname"  , x ) -> self#get_left.receptname  <- x ;
      | ("rightnodename"   , x ) -> self#get_right.nodename   <- x ;
      | ("rightreceptname" , x ) -> self#get_right.receptname <- x ;
      | ("kind"            , x ) -> () (* Constant field: cannot be set. *)

    (** Access method *)
    method is_connected =
      with_mutex mutex
        (fun () -> !connected);

    (** Make the cable connected, or do nothing if it's already connected: *)
    method private connect_right_now =
      with_mutex mutex
        (fun () ->
          (if not self#is_connected then begin
            Log.print_string ("Connecting the cable " ^ (self#get_name) ^ "\n");
            (* Turn on the relevant LEDgrid lights: *)
            let involved_devices_and_port_nos = self#involved_devices_and_port_nos in
            List.iter
              (fun (device, port_no) ->
                network#ledgrid_manager#set_port_connection_state
                  ~id:(device#id)
                  ~port:port_no
                  ~value:true
                  ())
              involved_devices_and_port_nos;
            connected := true;
            self#increment_alive_endpoints_no;
            Log.print_string "Ok: connected\n";
          end);
          refresh_sketch ());

    (** Make the cable disconnected, or do nothing if it's already disconnected: *)
    method private disconnect_right_now =
      with_mutex mutex
        (fun () ->
          (if self#is_connected then begin
            Log.print_string ("Disconnecting the cable " ^ (self#get_name) ^ "\n");
            (* Turn off the relevant LEDgrid lights: *)
            let involved_devices_and_port_nos = self#involved_devices_and_port_nos in
            List.iter
              (fun (device, port_no) ->
                network#ledgrid_manager#set_port_connection_state
                  ~id:(device#id)
                  ~port:port_no
                  ~value:false
                  ())
              involved_devices_and_port_nos;
            connected := false;
            self#decrement_alive_endpoints_no;
            Log.print_string "Ok: disconnected\n";
          end);
          refresh_sketch ());

   (** 'Suspending means disconnecting for cables *)
   method suspend_right_now =
     self#disconnect_right_now

   (** 'Resuming' means connecting for cables *)
   method resume_right_now =
     self#connect_right_now

   (** An always up-to-date 'reference counter' storing the number of alive
       endpoints plus the cable connection state (either 0 for 'disconnected' or
       1 for 'connected'). A cable can be started in the simulation when this is
       exactly 3, and must be terminated when it becomes less than 3. *)
   val alive_endpoints_no = ref 1 (* cables are 'connected' by default *)

   (** Check that the reference counter is in [0, 3]. To do: disable this for
   production. *)
   method private check_alive_endpoints_no =
    with_mutex mutex
      (fun () ->
        assert((!alive_endpoints_no >= 0) && (!alive_endpoints_no <= 3));
        Log.print_string "The reference count is now ";
        print_int !alive_endpoints_no; Log.print_string "\n")

   (** Record the fact that an endpoint has been created (at a lower level
       this means that its relevant {e hublet} has been created), and
       startup the simulated cable if appropriate. *)
   method increment_alive_endpoints_no =
     with_mutex mutex
       (fun () ->
         Log.print_string "\n+++ increment_alive_endpoints_no\n\n";
         self#check_alive_endpoints_no;
         alive_endpoints_no := !alive_endpoints_no + 1;
         self#check_alive_endpoints_no;
         if !alive_endpoints_no = 3 then begin
           Log.print_string "The reference count raised to three: starting up a cable\n";
           self#startup_right_now
         end)

   (** Record the fact that an endpoint is no longer running (at a lower level
       this means that its relevant {e hublet} has been destroyed), and
       shutdown the simulated cable if appropriate. *)
   method decrement_alive_endpoints_no =
     with_mutex mutex
       (fun () ->
         Log.print_string "\n--- decrement_alive_endpoints_no\n\n";
         self#check_alive_endpoints_no;
         alive_endpoints_no := !alive_endpoints_no - 1;
         self#check_alive_endpoints_no;
         if !alive_endpoints_no < 3 then begin
           (* Note that we destroy rather than terminating. This enables to re-create the
              simulated device later, at startup time, referring the correct hublets
              that will exist then, rather than the ones existing now *)
           Log.print_string "The reference count dropped below three: destroying a cable\n";
(*        self#destroy_right_now; (\* Before radical synchronization changes *\) *)
           self#destroy_right_now;
         end)

   (** Make a new simulated device according to the current status *)
   method private make_simulated_device =
     with_mutex mutex
       (fun () ->
         let left, right = self#get_left, self#get_right in
         let left_node = network#getNodeByName left.nodename in
         let right_node = network#getNodeByName right.nodename in
         let left_hublet_process =
           left_node#get_hublet_process_by_interface_name left.receptname in
         let right_hublet_process =
           right_node#get_hublet_process_by_interface_name right.receptname in
         let left_port_index =
           (left_node#getReceptacleByName left.receptname)#index in
         let right_port_index =
           (right_node#getReceptacleByName right.receptname)#index in
         let left_blink_command =
           match left_node#devkind with
             NotADevice -> None
           | _ -> Some (Printf.sprintf "(id: %i; port: %i)" left_node#id left_port_index) in
         let right_blink_command =
           match right_node#devkind with
             NotADevice -> None
           | _ -> Some (Printf.sprintf "(id: %i; port: %i)" right_node#id right_port_index) in
         Log.print_string ("left hublet process socket name is " ^ left_hublet_process#get_socket_name ^ "\n");
         Log.print_string ("right hublet process socket name is " ^ right_hublet_process#get_socket_name ^ "\n");
         new Simulated_network.ethernet_cable
           ~name:self#get_name
           ~left_end:left_hublet_process
           ~right_end:right_hublet_process
           ~blinker_thread_socket_file_name:(Some network#ledgrid_manager#blinker_thread_socket_file_name)
           ~left_blink_command
           ~right_blink_command
           ~unexpected_death_callback:self#destroy_because_of_unexpected_death
           ())

   (** This has to be overridden for cables, because we can't 'poweroff' as easily as the
       other devices: *)
   method private destroy_because_of_unexpected_death () =
     with_mutex mutex
       (fun () ->
         (* Refresh the process in some (ugly) way: *)
         if self#is_connected then begin
           (try self#disconnect_right_now with _ -> ());
           (try self#connect_right_now with _ -> ());
           connected := true;
         end else begin
           let current_alive_endpoints_no = !alive_endpoints_no in
           super_simulated_device#destroy_because_of_unexpected_death ();
           connected := true;
           alive_endpoints_no := 0;
           for i = 1 to current_alive_endpoints_no do
             self#increment_alive_endpoints_no;
           done
         end)

   (** To do: remove this ugly kludge, and make cables stoppable *)
   method can_startup = true (* To do: try reverting this *)
   method can_gracefully_shutdown = true (* To do: try reverting this *)
   method can_poweroff = true (* To do: try reverting this *)
   (** Only connected cables can be 'suspended' *)
   method can_suspend =
     with_mutex mutex
       (fun () -> !connected)
   (** Only non-connected cables with refcount exactly equal to 2 can be 'resumed' *)
   method can_resume =
     with_mutex mutex
       (fun () -> not !connected)

   (** A cable is never directly connected to any other cable *)
   method private get_involved_cables = []

   method receptacle_names =
    with_mutex mutex
      (fun () ->
        match self#socketnames with [ left_port; right_port ] ->
          let left_receptacle_name = left_port.receptname in
          let right_receptacle_name = right_port.receptname in
          left_receptacle_name, right_receptacle_name
        | _ -> assert false)

   method left_receptacle_name =
    with_mutex mutex
      (fun () ->
        let (left_receptacle_name, _) = self#receptacle_names in
        left_receptacle_name)

   method right_receptacle_name =
    with_mutex mutex
      (fun () ->
        let (_, right_receptacle_name) = self#receptacle_names in
        right_receptacle_name)

   method endpoints : (node * node) =
    with_mutex mutex
      (fun () ->
        let left_endpoint_name = self#get_left.nodename in
        let left_endpoint = network#getNodeByName left_endpoint_name in
        let right_endpoint_name = self#get_right.nodename in
        let right_endpoint = network#getNodeByName right_endpoint_name in
        left_endpoint, right_endpoint)

   method left_endpoint =
    with_mutex mutex
      (fun () ->
        let (left_endpoint, _) = self#endpoints in
        left_endpoint)

   method right_endpoint =
    with_mutex mutex
      (fun () ->
        let (_, right_endpoint) = self#endpoints in
        right_endpoint)

   (** Get the reference count right at the beginning: it starts at zero, but
       it's immediately incremented if endpoint hublet processes already
       exist: *)
   initializer
     let left_endpoint, right_endpoint = self#endpoints in
     self#set_correctness
       (network#would_a_cable_be_correct_between
          left_endpoint#name
          right_endpoint#name
          cablekind);
     (if left_endpoint#has_hublet_processes then
       self#increment_alive_endpoints_no);
     (if right_endpoint#has_hublet_processes then
       self#increment_alive_endpoints_no);
     Log.print_string ("The reference count for the just-created cable " ^ (self#get_name)^" is ");
     PervasivesExtra.print_int !alive_endpoints_no; Log.print_string "\n";
end;;

(** Function for make receptacles from 0 to k (so call it with desired number - 1) of desired portkind
    and with the given portprefix for name (for instance "eth"). *)
let rec mkrecepts network k portkind portprefix = match k with
| -1 -> []
| _ -> let r = new receptacle network (portprefix^(string_of_int k)) k nolabel portkind in
  (mkrecepts network (k-1) portkind portprefix)@[r]
;;


(* *************************** *
        class device
 * *************************** *)

let is_there_a_router_variant () =
  let result =
    List.mem "default" (MSys.variant_list_of ("router-" ^ router_unprefixed_filesystem) ())
  in
(*
  Log.printf "is_there_a_router_variant (): %b\n" result; flush_all ();
  List.iter
    (fun s -> Log.printf "* %s\n" s; flush_all ())
    (MSys.variant_list_of ("router-" ^ router_unprefixed_filesystem) ());
*)
  result;;

(** A device is a nodes s.t. nodekind=Device. Defects may be added after the creation. *)
class device =
  let variantsListOf d =
    MSys.variant_list_of ("router-" ^ d) () in
  let check_variant x d =
    if not (List.mem x (variantsListOf d)) then
      raise (Failure ("Setting router variant: "^x^" not in the list of available variants, which is:"^((variantsListOf d)=> StringExtra.Text.to_string)))
    else
      x in
 fun
  ~network
  ?(name="nodevicename")
  ?(label="")
  ?(devkind=Hub)
  ?(variant=no_variant_text)
  (ethnum:int) () ->
  let intialReceptacles = (mkrecepts network (ethnum-1) Eth "port") in

  object (self)

  inherit node ~network ~name ~label ~nodekind:Device ~devkind ~rlist:intialReceptacles () as super

  (** See the comment in the 'node' class for the meaning of this method: *)
  method polarity = if self#devkind = Router then MDI else MDI_X

  (** Get the full host pathname to the directory containing the guest hostfs
      filesystem: *)
  method hostfs_directory_pathname =
    if self#devkind = Router then
      match !simulated_device with
        Some d ->
          (d :> Simulated_network.router)#hostfs_directory_pathname
      | None ->
          failwith (self#name ^ " is not being simulated right now")
    else
      failwith "You can't use call hostfs_directory_pathname on a non-router device"

  (** Dot adjustments *)

  (** Returns an image representig the device with the given iconsize. *)
  method dotImg (z:iconsize) =
    let imgDir = (MSys.whereis "" MSys.Icon) in
    (imgDir^"ico."^(string_of_devkind self#devkind)^"."^(self#string_of_simulated_device_state)^"."^z^".png")

  (** Returns the label to use for cable representation.
      For devices, the port X is represented by the string "[X]". *)
  method dotLabelForEdges (receptname:string) =
    let index = (self#getReceptacleByName receptname)#index in
    ("["^string_of_int index^"]")

  (** Return the string representing the port in cable representation. *
      Ignore the receptname and returns the empty string. *)
  method dotPortForEdges (receptname:string) = ""

  method to_forest =
   Forest.leaf ("device",[
  		  ("name" , self#get_name) ;
                  ("label", self#get_label);
                  ("kind" , (string_of_devkind self#devkind)) ;
                  ("eth"  , (string_of_int (self#get_eth_number)));
                  ])

  (** A cable has just attributes (no childs) in this version. *)
  method eval_forest_attribute = function
  | ("name"  , x ) -> self#set_name  x
  | ("label" , x ) -> self#set_label x
  | ("kind"  , x ) -> self#set_devkind (devkind_of_string x)
  | ("eth"   , x ) -> self#set_eth_number (int_of_string x)

  (** Create the simulated device *)
  method private make_simulated_device =
    let id = self#id in
    let ethernet_receptacles = self#get_receptacles ~portkind:(Some Eth) () in
    if self#devkind != Router then
      let name = self#get_name in
      let hublets_no = List.length ethernet_receptacles in
      let unexpected_death_callback = self#destroy_because_of_unexpected_death in
      (match self#devkind with
        Hub -> new Simulated_network.hub
      | Switch -> new Simulated_network.switch
      | _ -> assert false)
        ~name ~hublets_no ~unexpected_death_callback ()
    else begin
      let cow_file_name =
        (Filesystem_history.get_states_directory ()) ^
        (Filesystem_history.add_state_for_device self#name) in
      Log.printf
        "About to start the router %s with the cow file %s\n"
        self#name
        cow_file_name;
      new Simulated_network.router
        ~name:self#get_name
        ~kernel_file_name:(MSys.marionnet_home_kernels ^ "linux-default")
        ~cow_file_name (* To do: use the variant!!! *)
        ~filesystem_file_name:(MSys.router_pathname_prefix^router_unprefixed_filesystem)
        ~ethernet_interfaces_no:(List.length ethernet_receptacles)
        ~umid:self#get_name
        ~id
        ~unexpected_death_callback:self#destroy_because_of_unexpected_death
        ()
    end

  (** Here we also have to manage LED grids: *)
  method private startup_right_now =
    (* Do as usual... *)
    super#startup_right_now;
    (* ...and also show the LED grid: *)
    network#ledgrid_manager#show_device_ledgrid ~id:(self#id) ();
(*
  (** Here we also have to manage LED grids and, for routers, cow files: *)
  method private gracefully_shutdown_right_now =
    (* Do as usual... *)
    super#gracefully_shutdown_right_now;
    (* ...and also hide the LED grid... *)
    network#ledgrid_manager#hide_device_ledgrid ~id:(self#id) ();
    (* ...and if this is a router then destroy, so that the next time we have to
       re-create a simulated device, and we start with a new cow *)
    if self#devkind = Router then
      self#destroy_right_now
*)
  (** Here we also have to manage cow files... *)
  method private gracefully_shutdown_right_now =
    (* Do as usual... *)
    super#gracefully_shutdown_right_now;
    (* ...and also hide the LED grid... *)
    network#ledgrid_manager#hide_device_ledgrid ~id:(self#id) ();
    (* Only for routers: we have to manage the hostfs stuff (when in exam mode) and
       destroy the simulated device, so that we can use a new cow file the next time: *)
    if self#devkind = Router then begin
      Log.printf "Calling hostfs_directory_pathname on %s...\n" self#name; flush_all ();
      let hostfs_directory_pathname = self#hostfs_directory_pathname in
      Log.printf "Ok, we're still alive\n"; flush_all ();
      (* If we're in exam mode then make the report available in the texts treeview: *)
      (if Command_line.are_we_in_exam_mode then begin
        let texts_interface = Texts_interface.get_texts_interface () in
        Log.printf "Adding the report on %s to the texts interface\n" self#name; flush_all ();
        texts_interface#import_report
          ~machine_or_router_name:self#name
          ~pathname:(hostfs_directory_pathname ^ "/report.html")
          ();
        Log.printf "Added the report on %s to the texts interface\n" self#name; flush_all ();
        (* (* We don't export the shell history for routers *)
        Log.printf "Adding the history on %s to the texts interface\n" self#name; flush_all ();
        texts_interface#import_history
          ~machine_or_router_name:self#name
          ~pathname:(hostfs_directory_pathname ^ "/bash_history.text")
          ();
        Log.printf "Added the history on %s to the texts interface\n" self#name; flush_all (); *)
      end);
      (* ...And destroy, so that the next time we have to re-create the process command line
         can use a new cow file (see the make_simulated_device method) *)
      self#destroy_right_now
    end;


  (** Here we also have to manage LED grids and, for routers, cow files: *)
  method private poweroff_right_now =
    (* Do as usual... *)
    super#poweroff_right_now;
    (* ...and also hide the LED grid... *)
    network#ledgrid_manager#hide_device_ledgrid ~id:(self#id) ();
    (* ...and if this is a router then destroy, so that the next time we have to
       re-create a simulated device, and we start with a new cow *)
    if self#devkind = Router then
      self#destroy_right_now

  (** A router has an associated initial variant *)
  val the_variant =
    ref (check_variant variant router_unprefixed_filesystem)

  method get_variant =
    if self#devkind = Router then
      !the_variant
    else
      super#get_variant (* fail *)

  method set_variant x =
    if self#devkind = Router then
      the_variant := check_variant x router_unprefixed_filesystem
    else
      failwith "Can not set a variant on a non-router device"
end;;


(* *************************** *
        class machine
 * *************************** *)

(** A machine is a node (a container of receptacles) s.t.
    nodekind=Machine. Some receptacles are immediatly added
    at creation time. *)
class machine =

  fun ~network
      ?(name="nomachinename")
      ?(label="")
      ?(mem:int=48)
      ?(ethnum:int=1)
      ?(distr:string="default")
      ?(variant:string=no_variant_text)
      ?(ker:string="default")
      ?(ter:string="X HOST") () ->

  (* Some checks over machine parameters *)

  let check_mem x = if x<8 or x>1024 then
    failwith ("Setting machine "^name^": value "^(string_of_int x)^" not in the memory range [8,1024]")
    else x in

  let check_ethnum x = if x<1 or x>5 then
    failwith ("Setting machine "^name^": value "^(string_of_int x)^" not in the eth's number range [1,5]")
    else x in

  let check_distrib x = if (not (x = "default")) && not (List.mem x (MSys.machine_filesystem_list ())) then
    failwith ("Setting machine "^name^": unknown GNU/Linux filesystem "^x)
    else x in

  let variantsListOf d =
    MSys.variant_list_of ("machine-" ^ d) () in

  let check_variant x d = if not (List.mem x (variantsListOf d)) then
    raise (Failure ("Setting machine "^name^": variant "^x^" not in the list of available variants, which is:"^((variantsListOf d)=> StringExtra.Text.to_string)))
    else x in

  let check_kernel x = if not (List.mem x (MSys.kernelList ())) then
    raise (Failure ("Setting machine "^name^": unknown kernel "^x))
    else x in

  let check_term x = if not (List.mem x (MSys.termList)) then
    raise (Failure ("Setting machine "^name^": unexpected uml terminal choice "^x))
    else x in

  let intialReceptacles =
   (mkrecepts network ((check_ethnum ethnum)-1) Eth "eth") in

  object (self)

  inherit node ~network ~name ~label ~nodekind:Machine ~rlist:intialReceptacles () as super

  (** See the comment in the 'node' class for the meaning of this method: *)
  method polarity = MDI

  (** Get the full host pathname to the directory containing the guest hostfs
      filesystem: *)
  method hostfs_directory_pathname =
    match !simulated_device with
      Some d ->
        (d :> Simulated_network.machine)#hostfs_directory_pathname
    | None ->
        failwith (self#name ^ " is not being simulated right now")

  (** Redefining node methods for adding checks *)
  method set_eth_number ?(prefix="eth") x = self#readjust_receptacleNumber ~prefix:(Some prefix) Eth (check_ethnum  x)

  (** A machine will be started with a certain amount of memory *)
  val mutable memory : int = check_mem mem

  method get_memory   = memory
  method set_memory x = memory <- check_mem x

  (** A machine has a Linux filesystem *)
  val mutable distrib : string = check_distrib distr

  method get_distrib   = distrib
  method set_distrib x = distrib <- check_distrib x

  (** A machine has an associated initial variant *)
  val mutable variant : string = check_variant variant distr

  method get_variant   = variant
  method set_variant x = variant <- check_variant x distrib

  (** A machine has a linux kernel. *)
  val mutable kernel : string = check_kernel ker

  method get_kernel   = kernel
  method set_kernel x = kernel <- check_kernel x

  (** A machine can be used accessed in a specific terminal mode. *)
  val mutable terminal : string = check_term ter

  method get_terminal   = terminal
  method set_terminal x = terminal <- check_term x

  (** Show for debugging *)
  method show = name

  (** Return an image representing the machine with the given iconsize. *)
  method dotImg (z:iconsize) =
    let imgDir = (MSys.whereis "" MSys.Icon) in
    (imgDir^"ico.machine."^(self#string_of_simulated_device_state)^"."^z^".png") (* distinguer redhat,debian, etc ? *)

  (** Machine to forest encoding. *)
  method to_forest =
   Forest.leaf ("machine", [
                   ("name"     ,  self#get_name ) ;
                   ("memory"   ,  (string_of_int self#get_memory)) ;
                   ("distrib"  ,  self#get_distrib  ) ;
                   ("variant"  ,  self#get_variant  ) ;  (* EX "state" attenzione nel from_forest *)
                   ("kernel"   ,  self#get_kernel   ) ;
                   ("terminal" ,  self#get_terminal ) ;
                   ("eth"      ,  (string_of_int self#get_eth_number))  ;
	           ])

 (** A machine has just attributes (no childs) in this version. *)
 method eval_forest_attribute = function
  | ("name"     , x ) -> self#set_name x
  | ("memory"   , x ) -> self#set_memory (int_of_string x)
  | ("distrib"  , x ) -> self#set_distrib x
  | ("variant"  , x ) -> self#set_variant x
  | ("kernel"   , x ) -> self#set_kernel x
  | ("terminal" , x ) -> self#set_terminal x
  | ("eth"      , x ) -> self#set_eth_number  (int_of_string x)
  | _ -> () (* Forward-comp. *)


  (** Create the simulated device *)
  method private make_simulated_device =
    let id = self#id in
    let ethernet_receptacles = self#get_receptacles ~portkind:(Some Eth) () in
    let cow_file_name =
      (Filesystem_history.get_states_directory ()) ^
      (Filesystem_history.add_state_for_device self#name) in
    Log.printf
      "About to start the machine %s with the cow file %s\n"
      self#name
      cow_file_name;
    new Simulated_network.machine
      ~name:self#get_name
      ~kernel_file_name:(MSys.kernel_pathname_prefix ^ kernel)
      ~filesystem_file_name:(MSys.machine_pathname_prefix^distrib)
      ~cow_file_name
      ~ethernet_interfaces_no:(List.length ethernet_receptacles)
      ~memory:self#get_memory
      ~umid:self#get_name
      ~id
      ~xnest:(match MSys.x_policy_of_string terminal with MSys.Xnest -> true | _ -> false)
      ~unexpected_death_callback:self#destroy_because_of_unexpected_death
      ()

  (** Here we also have to manage cow files... *)
  method private gracefully_shutdown_right_now =
    Log.printf "Calling hostfs_directory_pathname on %s...\n" self#name; flush_all ();
    let hostfs_directory_pathname = self#hostfs_directory_pathname in
    Log.printf "Ok, we're still alive\n"; flush_all ();
    (* Do as usual... *)
    super#gracefully_shutdown_right_now;
    (* If we're in exam mode then make the report available in the texts treeview: *)
    (if Command_line.are_we_in_exam_mode then begin
      let texts_interface = Texts_interface.get_texts_interface () in
      Log.printf "Adding the report on %s to the texts interface\n" self#name; flush_all ();
      texts_interface#import_report
        ~machine_or_router_name:self#name
        ~pathname:(hostfs_directory_pathname ^ "/report.html")
        ();
      Log.printf "Added the report on %s to the texts interface\n" self#name; flush_all ();
      Log.printf "Adding the history on %s to the texts interface\n" self#name; flush_all ();
      texts_interface#import_history
        ~machine_or_router_name:self#name
        ~pathname:(hostfs_directory_pathname ^ "/bash_history.text")
        ();
      Log.printf "Added the history on %s to the texts interface\n" self#name; flush_all ();
    end);
    (* ...And destroy, so that the next time we have to re-create the process command line
       can use a new cow file (see the make_simulated_device method) *)
    self#destroy_right_now

  (** Here we also have to manage cow files... *)
  method private poweroff_right_now =
    (* Do as usual... *)
    super#poweroff_right_now;
    (* ...And destroy, so that the next time we have to re-create the process command line
       can use a new cow file (see the make_simulated_device method) *)
    self#destroy_right_now

end;;


(* *************************** *
        class cloud
 * *************************** *)

(** A cloud is an unknown network with some gateways. *)
class cloud =
  fun ~network
      ?(name="nocloudname")
      ?(label="")
      () ->

  let gwnum             = 2 in
  let intialReceptacles = (mkrecepts network (gwnum-1) Eth "port") in

  object (self)

  inherit node ~network ~name ~label ~nodekind:Cloud ~rlist:intialReceptacles ()

  (** See the comment in the 'node' class for the meaning of this method: *)
  method polarity = MDI_X

  method show = self#name

  method eth_number        = 2

  (** Dot adjustments *)

  (** Return an image representing the cloud with the given iconsize. *)
  method dotImg (z:iconsize) =
    let imgDir = (MSys.whereis "" MSys.Icon) in
    (imgDir^"ico.cloud."^(self#string_of_simulated_device_state)^"."^z^".png")

  (** Cloud endpoints are represented in the same way of devices ones, with "[X]". *)
  method dotLabelForEdges (receptname:string) =
    let index = (self#getReceptacleByName receptname)#index in
    ("["^string_of_int index^"]")

  method dotPortForEdges (receptname:string) = ""

  method to_forest =
   Forest.leaf ("cloud",[
  		 ("name",self#get_name);
  		 ])

 (** A cloud has just attributes (no childs) in this version. *)
 method eval_forest_attribute = function
  | ("name"     , x ) -> self#set_name x

  (** Create the simulated device *)
  method private make_simulated_device =
    new Simulated_network.cloud
      ~name:self#get_name
      ~unexpected_death_callback:self#destroy_because_of_unexpected_death
(*       ~id *)
      ()
end;; (* cloud *)


(* *************************** *
        class gateway
 * *************************** *)

(** An internet gateway is simply a node with a single receptacle and an associated IP number. *)
class gateway =

  fun ~network
      ?(name="nogatewayname")
      ?(label="")
      () ->

  let intialReceptacles = (mkrecepts network 0 Eth "eth") in

  object (self)

  inherit node ~network ~name ~label ~nodekind:Gateway ~rlist:intialReceptacles ()

  (** See the comment in the 'node' class for the meaning of this method: *)
  method polarity = MDI_X

  method show = (self#name^" (gateway)")

  (** Dot adjustments *)

  (** Return an image representing the machine with the given iconsize. *)
  method dotImg (z:iconsize) =
    let imgDir = (MSys.whereis "" MSys.Icon) in
    (imgDir^"ico.socket."^(self#string_of_simulated_device_state)^"."^z^".png")

  (** Returns the label to use for cable representation.*)
  method dotLabelForEdges receptname = "" (* ip#toString *)

  (** Returns the port to use for cable representation.
      Ignore the receptname and returns the empty string. *)
  method dotPortForEdges receptname = ""

  method to_forest =
   Forest.leaf ("gateway", [
                  ("name", self#get_name);
                  ])

  (** A gateway has just attributes (no childs) in this version. *)
  method eval_forest_attribute = function
  | ("name"     , x ) -> self#set_name x

  (** Create the simulated device *)
  method private make_simulated_device =
    new Simulated_network.gateway
      ~name:self#get_name
      ~bridge_name:Global_options.ethernet_socket_bridge_name
      ~unexpected_death_callback:self#destroy_because_of_unexpected_death
      ()
end;; (* gwinternet *)


(*********************
    Submodule Edge
 *********************)

module Edge = struct

(** Type of an edge. An edge is a triple (node1,receptname1), cable , (node2,receptname2) *)
type edge = (node*string) * cable * (node*string)
;;

(** The Dot translation of an edge of the graph. *)
let dotEdgeTrad ?(edgeoptions="") (labeldistance_base) (((n1,r1),c,(n2,r2)):edge) =
    begin

    let vertexlab node iden recept =

      let port      = node#dotPortForEdges  recept in
      let portlabel = node#dotLabelForEdges recept in

      match port,portlabel with
      | "",""  -> ("")
      | _ , "" -> (","^iden^"=\""^port^"\"")
      | _ , _  ->
          begin
          let port_line      = (StringExtra.assemble "<TR><TD>" port "</TD></TR>") in
          let portlabel_line = (StringExtra.assemble "<TR><TD><FONT COLOR=\"#3a3936\">" portlabel "</FONT></TD></TR>") in
(","^iden^"=<
<TABLE BORDER=\"0\" CELLBORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"0\">
"^port_line^"
"^portlabel_line^"
</TABLE>
>")       end in


   let labeldistance =
      begin
      let p1  = n1#dotPortForEdges  r1 in
      let pl1 = n1#dotLabelForEdges r1 in
      let p2  = n2#dotPortForEdges  r2 in
      let pl2 = n2#dotLabelForEdges r2 in

      (* if there is a vertex with both port and portlabel not empty => set labeldistance +0.5 *)
      if ((p1<>"" && pl1<>"") or (p2<>"" && pl2<>""))
      then ("labeldistance="^(string_of_float (labeldistance_base +. 0.5))^",") else ""
      end in


   let (tail, head, taillabel, headlabel) =

    (* Invert left and right sides of the cable if required *)
    let (n1,r1,n2,r2) = if c#dotoptions#reverted#get then (n2,r2,n1,r1) else (n1,r1,n2,r2) in

     match (n1#nodekind, c#cablekind, n2#nodekind, n1#get_label, n2#get_label) with

     |   _   ,    _   ,    _  , "", "" -> (n1#name^":img"), (n2#name^":img"), (vertexlab n1 "taillabel" r1), (vertexlab n2 "headlabel" r2)
     |   _   ,    _   ,    _  , "", l2 -> (n1#name^":img"), (n2#name)       , (vertexlab n1 "taillabel" r1), (vertexlab n2 "headlabel" r2)
     |   _   ,    _   ,    _  , l1, "" -> (n1#name)       , (n2#name^":img"), (vertexlab n1 "taillabel" r1), (vertexlab n2 "headlabel" r2)
     |   _   ,    _   ,    _  , l1, l2 -> (n1#name)       , (n2#name)       , (vertexlab n1 "taillabel" r1), (vertexlab n2 "headlabel" r2)

   in

   let edgeoptions = if edgeoptions = "" then "" else (edgeoptions^",") in
   let cable_label = c#get_name ^ (if c#get_label = "" then "" else ("  "^c#get_label)) in

   let edgeoptions = edgeoptions ^ "arrowhead=obox, arrowtail=obox, arrowsize=0.4," ^
    (if c#is_connected then ""
   		       else "style=dashed,") in

   let label_color = if c#cablekind = Direct then Color.direct_cable else Color.crossover_cable in

    (tail^" -> "^head^" ["^edgeoptions^labeldistance^"label=<<FONT COLOR=\""^label_color^"\">"^cable_label^"</FONT>>"^taillabel^headlabel^"];")

 end;; (* function dotEdgeTrad *)


end;; (* Module Edge *)


open Edge;;


(* *************************** *
        class network
 * *************************** *)

(** Class modelling the virtual network *)
class network () =

 object (self)
 inherit Xforest.interpreter ()

 (** Motherboard is set in Gui_motherboard. *)
 val mutable motherboard : State_types.motherboard option = None
 method motherboard = match motherboard with Some x -> x | None -> assert false
 method set_motherboard m = motherboard <- Some m

 val mutable machines : (machine list) = []
 val mutable devices  : (device  list) = []
 val mutable cables   : (cable   list) = []
 val mutable clouds   : (cloud   list) = []
 val mutable gateways : (gateway list) = []

 val ledgrid_manager = Ledgrid_manager.the_one_and_only_ledgrid_manager

 (** Buffers to backup/restore data. *)
 val mutable machines_buffer : (machine list) = []
 val mutable devices_buffer  : (device  list) = []
 val mutable cables_buffer   : (cable   list) = []
 val mutable clouds_buffer   : (cloud   list) = []
 val mutable gateways_buffer : (gateway list) = []

 (** Accessors *)

 method machines        = machines
 method devices         = devices
 method cables          = cables
 method clouds          = clouds
 method gateways        = gateways
 method ledgrid_manager = ledgrid_manager

 (** Related dot options fro drawing this virtual network.
     This pointer is shared with the project instance. *)
 val mutable dotoptions : (Dotoptions.network option) = None
 method      dotoptions   = match dotoptions with Some x -> x | None -> raise (Failure "network#dotoptions")
 method  set_dotoptions x = dotoptions <- Some x

 method components : (component list) =
   ((machines :> component list) @
    (devices  :> component list) @
    (clouds   :> component list) @
    (gateways :> component list) @
    (cables   :> component list)    (* CABLES MUST BE AT THE FINAL POSITION for marshaling !!!! *)
    )

 (** Setter *)

 method reset () =
   Log.print_string "reset: begin\n";
   Log.print_string "resetting the LEDgrid manager...\n";
   Log.print_string "destroying all cables...\n";
   (List.iter
      (fun cable -> try cable#destroy with _ -> ()) (* "right_now" was here before the recent radical synchronization changes*)
      cables);
   Log.print_string "destroying all machines...\n";
   (List.iter
      (fun machine -> try machine#destroy with _ -> ()) (* "right_now" was here before the recent radical synchronization changes*)
      machines);
   Log.print_string "destroying all switchs, hubs and routers...\n";
   (List.iter
      (fun device -> try device#destroy with _ -> ()) (* "right_now" was here before the recent radical synchronization changes*)
      devices);
   Log.print_string "destroying all clouds...\n";
   (List.iter
      (fun cloud -> try cloud#destroy with _ -> ()) (* "right_now" was here before the recent radical synchronization changes*)
      clouds);
   Log.print_string "destroying all gateways...\n";
   (List.iter
      (fun gateway -> try gateway#destroy with _ -> ()) (* "right_now" was here before the recent radical synchronization changes*)
      gateways);
   Log.print_string "Synchronously wait that everything terminates...\n";
   Task_runner.the_task_runner#wait_for_all_currently_scheduled_tasks;

   Log.print_string "Making the network graph empty...\n";
   machines <- [] ;
   devices  <- [] ;
   clouds   <- [] ;
   gateways <- [] ;
   cables   <- [] ;

   Log.print_string "Wait for all devices to terminate...\n"; flush_all ();
   (** Make sure that all devices have actually been terminated before going
       on: we don't want them to lose filesystem access: *)
   Log.print_string "All devices did terminate.\n"; flush_all ();

   Log.print_string "reset: end (success)\n";

 method restore_from_buffers =
  begin
   self#reset ();
   machines <- machines_buffer ;
   devices  <- devices_buffer  ;
   clouds   <- clouds_buffer   ;
   gateways <- gateways_buffer ;
   cables   <- cables_buffer
 end

 method save_to_buffers =
  begin
   machines_buffer <- machines ;
   devices_buffer  <- devices  ;
   clouds_buffer   <- clouds   ;
   gateways_buffer <- gateways ;
   cables_buffer   <- cables
  end

 method to_forest =
   let l = List.map (fun x->x#to_forest) self#components in
   Forest.tree ("network",[]) (Forest.of_treelist l)

 (** We redefine just the interpretation of a childs.
     We ignore (in this version) network attributes. *)
 method eval_forest_child (f:Xforest.tree) : unit = try begin
  match f with
  | Forest.NonEmpty (("machine", attrs) , childs , Forest.Empty) ->
      let x = new machine ~network:self () in
      x#from_forest ("machine", attrs) childs ;
      self#addMachine x

 | Forest.NonEmpty (("device", attrs) , childs , Forest.Empty) ->
      let ethno = List.assoc "eth" attrs in
      let x = new device ~network:self (int_of_string ethno) () in
      x#from_forest ("device", attrs) childs ;
      self#addDevice x

 | Forest.NonEmpty (("cloud", attrs) , childs , Forest.Empty) ->
      let x = new cloud ~network:self () in
      x#from_forest ("cloud", attrs) childs ;
      self#addCloud x;

 | Forest.NonEmpty (("gateway", attrs) , childs , Forest.Empty) ->
      let x = new gateway ~network:self () in
      x#from_forest ("gateway", attrs) childs  ;
      self#addGateway x;

 | Forest.NonEmpty (("cable", attrs) , childs , Forest.Empty) ->
      (* Cables represent a special case: they must be builded knowing their endpoints. *)
      let ln = List.assoc "leftnodename"    attrs in
      let lr = List.assoc "leftreceptname"  attrs in
      let rn = List.assoc "rightnodename"   attrs in
      let rr = List.assoc "rightreceptname" attrs in
      let ck = List.assoc "kind"            attrs in
      let left  = { nodename=ln; receptname=lr }  in
      let right = { nodename=rn; receptname=rr }  in
      let cablekind = cablekind_of_string ck      in
      let x = new cable ~motherboard:self#motherboard ~cablekind ~network:self ~left ~right () in
      x#from_forest ("cable", attrs) childs ;
      self#addCable x;

  | Forest.NonEmpty ((nodename, _) , _ , _)
     -> (Log.printf "network#eval_forest_child: I can't interpret this nodename '%s'.\n" nodename)
        (* Forward-compatibility *)

  | Forest.Empty
     -> (Log.printf "network#eval_forest_child: I can't interpret the empty forest.\n")
        (* Forward-compatibility *)
 end
 with e  ->
   Log.printf "network#eval_forest_child: something goes wrong interpreting a child (%s).\n"
   (Printexc.to_string e); () (* Forward-compatibility *)

 (* Destruct the x-value into a concrete xml string *)
(* method xml : string =
    let trad = ref "" in
    let update x = trad:=!trad ^ x in
    let _ = Ocamlduce.Print.print_xml update self#xvalue in
    let s = Xml.parse_string !trad in
    Printf.sprintf "%s" (Xml.to_string_fmt s)*)

 method nodes : (node list) =
   ((machines :> node list) @ (devices  :> node list) @
    (clouds   :> node list) @ (gateways :> node list) )

 method names = (List.map (fun x->x#name) self#components)

 method suggestedName prefix =
   let rec tip prefix k =
     begin
     let prop = (prefix^(string_of_int k)) in
     if self#nameExists prop then tip prefix (k+1) else prop
     end in tip prefix 1


 (** Get machine, device or cable in the network by its name *)
 method getMachineByName n = try List.find (fun x->x#name=n) machines   with _ -> failwith ("getMachineByName "^n)
 method getDeviceByName  n = try List.find (fun x->x#name=n) devices    with _ -> failwith ("getDeviceByName "^n)
 method getCableByName   n = try List.find (fun x->x#name=n) cables     with _ -> failwith ("getCableByName "^n)
 method getCloudByName   n = try List.find (fun x->x#name=n) clouds     with _ -> failwith ("getCloudByName "^n)
 method getGatewayByName n = try List.find (fun x->x#name=n) gateways   with _ -> failwith ("getGatewayByName "^n)
 method getNodeByName    n = try List.find (fun x->x#name=n) self#nodes with _ -> failwith ("getNodeByName \""^n^"\"")

 (** Managing socketnames *)

 (** List of socketnames involved in the network (in its nodes) *)
 method socketnames = List.flatten (List.map (fun node->node#socketnames) (self#nodes))

 (** List of socketnames busy by a cable in the network *)
 method busySocketnames = List.flatten (List.map (fun c->c#socketnames) cables)

 (** List of available socketnames in the network *)
 method freeSocketnames = ListExtra.substract self#socketnames self#busySocketnames

 (** Is a socketname available for connection? *)
 method isSocketnameFree x = List.mem x self#freeSocketnames

 (** Managing receptacle of a given node *)

 (** List of receptacles of a node and of a given kind. *)
 method receptaclesOfNode name (k:portkind) =
     (self#getNodeByName name)#get_receptacles ~portkind:(Some k) ()

 (** List of receptacles names of a node and of a given kind. *)
 method receptaclesNamesOfNode name (k:portkind) =
     (self#getNodeByName name)#receptacleNames ~portkind:(Some k) ()

 (** List of free receptacles of a nodename and of a given portkind. *)
 method freeReceptaclesOfNode nodename portkind =
   let recepts = (self#receptaclesOfNode nodename portkind) in
   let freeSocketnames  = self#freeSocketnames in
   List.filter (fun r ->(List.mem {nodename=nodename;receptname=r#name} freeSocketnames)) recepts

 (** List of free receptacles names of a nodename and of a given portkind. *)
 method freeReceptaclesNamesOfNode nodename portkind =
   let recept_names   = (self#receptaclesNamesOfNode nodename portkind) in
   let freeSocketnames  = self#freeSocketnames in
   List.filter (fun r ->(List.mem {nodename=nodename;receptname=r} freeSocketnames)) recept_names

 (** List of busy receptacles of a nodename and of a given portkind. *)
 method busyReceptaclesOfNode nodename portkind =
   let recepts = (self#receptaclesOfNode nodename portkind) in
   let freeSocketnames  = self#freeSocketnames in
   List.filter (fun r ->(not (List.mem {nodename=nodename;receptname=r#name} freeSocketnames))) recepts

 (** List of busy receptacles names of a nodename and of a given portkind. *)
 method busyReceptaclesNamesOfNode nodename portkind =
   let recept_names   = (self#receptaclesNamesOfNode nodename portkind) in
   let freeSocketnames  = self#freeSocketnames in
   List.filter (fun r ->not (List.mem {nodename=nodename;receptname=r} freeSocketnames)) recept_names

 (** The max index among busy receptacles of a given kind of a given node.
     The user cannot change the number of receptacle of the given node to a number less than this index+1.
     For instance, if the (maxBusyReceptacleIndex "rome" Eth) = 2 then the user can change
     the number of receptacle of rome but only with a number >= 3.  *)
 method maxBusyReceptacleIndex nodename portkind : int =
   let bl = (self#busyReceptaclesOfNode nodename portkind) in
   let il = (List.map (fun x->x#index) bl) in
   if il=[] then -1 else ListExtra.max il

 (** Component exists? *)

 method machineExists n = let f=(fun x->x#name=n) in (List.exists f machines)
 method deviceExists  n = let f=(fun x->x#name=n) in (List.exists f devices )
 method cableExists   n = let f=(fun x->x#name=n) in (List.exists f cables  )
 method cloudExists   n = let f=(fun x->x#name=n) in (List.exists f clouds  )
 method gatewayExists n = let f=(fun x->x#name=n) in (List.exists f gateways)
 method nameExists    n = List.mem n self#names

 (** What kind of node is it ? *)
 method nodeKind n = (self#getNodeByName n)#nodekind

 (** Adding components *)

   method make_device_ledgrid (d:device) =
     (* Make a new device LEDgrid: *)
     self#ledgrid_manager#make_device_ledgrid
       ~id:(d#id)
       ~title:(d#get_name(*  ^ " (" ^(string_of_devkind d#devkind) ^ ")" *))
       ~label:(match d#devkind with
                 Hub -> "Hub"
               | Switch -> "Switch"
               | Router -> "Router"
               | _ -> assert false)
       ~ports_no:(d#numberOfReceptacles ~portkind:(Some Eth) ())
       ~image_directory:(MSys.marionnet_home_images^"/leds/"^(string_of_devkind d#devkind))
       ();
     (* Set port connection state: *)
     let busy_ports =
       self#busyReceptaclesOfNode d#get_name Eth in
     let busy_ports_as_indices =
       List.map (fun receptacle -> receptacle#index) busy_ports in
     ignore (List.map
               (fun port_index -> (self#ledgrid_manager#set_port_connection_state
                                     ~id:d#id
                                     ~port:port_index
                                     ~value:true
                                     ()))
               busy_ports_as_indices)

 (** Devices must have a unique name in the network *)
 method addDevice (d:device) =
    if (self#nameExists d#name) then
      raise (Failure "addDevice : name already used in the network")
    else begin
      devices  <- (devices@[d]);
      self#make_device_ledgrid d;
      if (d#devkind = Router) &&
         ((Filesystem_history.number_of_states_with_name d#name) = 0) then (* not after load *)
        Filesystem_history.add_device
          d#name
          (MSys.router_prefix ^ router_unprefixed_filesystem)
          d#get_variant
          "router"
    end

 (** Machines must have a unique name in the network *)
 method addMachine (m:machine) =
   if (self#nameExists m#name) then
     raise (Failure ("addMachine : name "^m#name^" already used in the network"))
   else begin
     (* Ok, we have fixed the variant name. Now just prepend the machine to the
        appropriate list: *)
     machines  <- (machines@[m]);
   end
 (** Cable must connect free socketnames  *)
 method addCable (c:cable) =
    if (self#nameExists c#name)
    then raise (Failure "addCable : name already used in the network")
    else
      if (self#isSocketnameFree c#get_left) && (self#isSocketnameFree c#get_right)
      then begin
        cables  <- (cables@[c]);
        (* If at least one endpoint is a device then set the port state to connected in
           the appropriate LED grid: *)
        let left_endpoint, right_endpoint = c#endpoints in
        let left_port_name, right_port_name = c#receptacle_names in
        let left_port_index = (left_endpoint#getReceptacleByName left_port_name)#index in
        let right_port_index = (right_endpoint#getReceptacleByName right_port_name)#index in
(*         Log.print_string ("\nLeft side: " ^ left_port_name ^ "\n Right side: "^right_port_name^"\n"); *)
        if left_endpoint#devkind != NotADevice then
          self#ledgrid_manager#set_port_connection_state
            ~id:(left_endpoint#id)
            ~port:left_port_index
            ~value:true
            ();
        if right_endpoint#devkind != NotADevice then
          self#ledgrid_manager#set_port_connection_state
            ~id:(right_endpoint#id)
            ~port:right_port_index
            ~value:true
            ();
      end else
        raise (Failure ("addCable : left and/or right socketname busy or nonexistent for "^(c#show "")))

 (** Clouds must have a unique name in the network *)
 method addCloud (c:cloud) =
    if (self#nameExists c#name)
          then raise (Failure ("addCloud : name "^c#name^" already used in the network"))
          else clouds  <- (clouds@[c])

 (** Gateways must have a unique name in the network *)
 method addGateway (g:gateway) =
    if (self#nameExists g#name)
          then raise (Failure ("addGateway : name "^g#name^" already used in the network"))
          else gateways  <- (gateways@[g])


 (** Removing components *)

 (** Remove a machine from the network. Remove it from the [machines] list and remove all related cables *)
 method delMachine mname =
     let m  = self#getMachineByName mname in
     (* Destroy cables first, from the network and from the defects treeview (cables are not
        in the network details treeview): they refer what we're removing... *)
     let cables_to_remove = List.filter (fun c->c#isNodeinvolved mname) cables in
     let defects = Defects_interface.get_defects_interface () in
     List.iter
       (fun cable ->
         defects#remove_cable cable#name;
         self#delCable cable#name)
       cables_to_remove;
     Filesystem_history.remove_device_tree mname;
     m#destroy;(*      m#destroy_right_now; *) (* This was here before the radical synchronization changes *)
     machines  <- List.filter (fun x->not (x=m)) machines

 (** Remove a device from the network. Remove it from the [devices] list and remove all related cables *)
 method delDevice dname =
     let d  = self#getDeviceByName dname in
     (* Destroy cables first: they refer what we're removing... *)
     let cables_to_remove = List.filter (fun c->c#isNodeinvolved dname) cables in
     let defects = Defects_interface.get_defects_interface () in
     List.iter
       (fun cable ->
         defects#remove_cable cable#name;
         self#delCable cable#name)
       cables_to_remove;
     (if d#devkind = Router then
       Filesystem_history.remove_device_tree dname);
     d#destroy;(*      d#destroy_right_now; *) (* This was here before the radical synchronization changes *)
     devices  <- List.filter (fun x->not (x=d)) devices

 (** Remove a cable from network *)
 method delCable cname =
     let c = self#getCableByName cname in
(*      cables  <- (cables@[c]); *) (* No, this was a cut/paste error, I think. *)
     (* If at least one endpoint is a device then set the port state to disconnected in
        the appropriate LED grid: *)
     let left_endpoint, right_endpoint = c#endpoints in
     let left_port_name, right_port_name = c#receptacle_names in
     let left_port_index = (left_endpoint#getReceptacleByName left_port_name)#index in
     let right_port_index = (right_endpoint#getReceptacleByName right_port_name)#index in
(*         Log.print_string ("\nLeft side: " ^ left_port_name ^ "\n Right side: "^right_port_name^"\n"); *)
     (if left_endpoint#devkind != NotADevice then
       self#ledgrid_manager#set_port_connection_state
         ~id:(left_endpoint#id)
         ~port:left_port_index
         ~value:false
         ());
     (if right_endpoint#devkind != NotADevice then
       self#ledgrid_manager#set_port_connection_state
         ~id:(right_endpoint#id)
         ~port:right_port_index
         ~value:false
         ());
     cables <- (List.filter (fun x->not (x=c)) cables);
(* To do: c#destroy_right_now was at the beginning, but I think it should be here. *)
     c#destroy;(*      c#destroy_right_now; *) (* This was here before the radical synchronization changes *)
     (* Remove the cable also from the defects interface (cables are not in the network
        details interface): *)
     (* (* This should not be needed; it's actually harmful *)
     (try
       let defects = Defects_interface.get_defects_interface () in
       defects#remove_cable cname
     with _ ->
       ()); *)

 (** Remove a cloud from the network. Remove it from the [clouds] list and remove all related cables *)
 method delCloud clname =
     let cl  = self#getCloudByName clname in
     (* Destroy cables first: they refer what we're removing... *)
     let cables_to_remove = List.filter (fun c->c#isNodeinvolved clname) cables in
     let defects = Defects_interface.get_defects_interface () in
     List.iter
       (fun cable ->
         defects#remove_cable cable#name;
         self#delCable cable#name)
       cables_to_remove;
     cl#destroy;(*      cl#destroy_right_now; *) (* This was here before the radical synchronization changes *)
     clouds  <- List.filter (fun x->not (x=cl)) clouds

 (** Remove a gateway from the network. Remove it from the [gateways] list and remove all related cables *)
 method delGateway gname =
     let g  = self#getGatewayByName gname in
     (* Destroy cables first: they refer what we're removing... *)
     let cables_to_remove = List.filter (fun c->c#isNodeinvolved gname) cables in
     let defects = Defects_interface.get_defects_interface () in
     List.iter
       (fun cable ->
         defects#remove_cable cable#name;
         self#delCable cable#name)
       cables_to_remove;
     g#destroy;(*      g#destroy_right_now; *) (* This was here before the radical synchronization changes *)
     gateways  <- List.filter (fun x->not (x=g)) gateways

 (** Change the name of a node => change all cable socketnames refering this node *)
 method changeNodeName oldname newname =
   if oldname = newname then () else
   let node = self#getNodeByName oldname in
   node#set_name newname ;
   ListExtra.foreach cables (fun c->c#changeNodeName oldname newname)

 (** Facilities *)

  (** Would a hypothetical cable of a given crossoverness be 'correct' if it connected the two
      given nodes? *)
  method would_a_cable_be_correct_between endpoint1_name endpoint2_name crossoverness =
    let polarity1 = (self#getNodeByName endpoint1_name)#polarity in
    let polarity2 = (self#getNodeByName endpoint2_name)#polarity in
    (* We need a crossover cable if the polarity is the same: *)
    match polarity1, polarity2 with
     | Intelligent , _      | _           , Intelligent -> true
     | MDI_X       , MDI    | MDI         , MDI_X       -> crossoverness = Direct
     | MDI_X       , MDI_X  | MDI         , MDI         -> crossoverness = Crossover

 (** List of machines names in the network *)
 method getMachineNames     =
   List.map (fun x->x#name) machines

 (** List of device names in the network *)
 method getDeviceNames     =
   List.map (fun x->x#name) devices

 (** List of node names in the network *)
 method getNodeNames  =
   List.map (fun x->x#name) (self#nodes)

 (** List of hub names in the network *)
 method getHubNames     =
   let hublist= List.filter (fun x->x#devkind=Hub) devices in
   List.map (fun x->x#name) hublist

 (** List of switch names in the network *)
 method getSwitchNames  =
   let swlist= List.filter (fun x->x#devkind=Switch) devices in
   List.map (fun x->x#name) swlist

 (** List of router names in the network *)
 method getRouterNames  =
   let routerlist= List.filter (fun x->x#devkind=Router) devices in
   List.map (fun x->x#name) routerlist

 (** List of direct cable names in the network *)
 method getDirectCableNames  =
   let clist= List.filter (fun x->x#cablekind=Direct) cables in
   List.map (fun x->x#name) clist

 (** List of crossover cable names in the network *)
 method getCrossoverCableNames  =
   let clist= List.filter (fun x->x#cablekind=Crossover) cables in
   List.map (fun x->x#name) clist

 (** List of serial cable names in the network *)
 method getSerialCableNames  =
   let clist= List.filter (fun x->x#cablekind=NullModem) cables in
   List.map (fun x->x#name) clist

 (** List of cloud names in the network *)
 method getCloudNames     =
   List.map (fun x->x#name) clouds

 (** List of gateway names in the network *)
 method getGatewayNames     =
   List.map (fun x->x#name) gateways

 (** Starting and showing the network *)

 (** List of reverted cables (used only for drawing network) *)
 method revertedCables : (string list) =
   let clist= List.filter (fun x->x#dotoptions#reverted#get) cables in
   List.map (fun x->x#name) clist

 (** Set the reverted dotoptions field of a cable of the network (identified by name) *)
 method revertedCableSet (x:bool) (cname:string) =
   (self#getCableByName cname)#dotoptions#reverted#set x

 (** Show network topology *)
 method show =
   Log.print_endline "========== NETWORK STATUS ===========";
   (* show devices *)
   let msg= try
        (StringExtra.Fold.commacat
        (List.map (fun d->d#name^" ("^(string_of_devkind d#devkind)^")") devices))
        with _ -> ""
     in Log.print_endline ("Devices \r\t\t: "^msg);
   (* show machines *)
   let msg=try
        (StringExtra.Fold.commacat (List.map (fun m->m#show) machines))
        with _ -> ""
   in Log.print_endline ("Machines \r\t\t: "^msg);
   (* show clouds *)
   let msg=try
        (StringExtra.Fold.commacat (List.map (fun c->c#show) clouds))
        with _ -> ""
   in Log.print_endline ("Clouds \r\t\t: "^msg);
   (* show gateways *)
   let msg=try
        (StringExtra.Fold.commacat (List.map (fun c->c#show) gateways))
        with _ -> ""
   in Log.print_endline ("Gateways \r\t\t: "^msg);
   (* show links *)
   let msg=try
        (StringExtra.Fold.newlinecat (List.map (fun c->(c#show "\r\t\t  ")) cables))
        with _ -> ""
   in Log.print_endline ("Cables \r\t\t: "^msg)


 (** {b Consider cable as Edge.edges} *)

 (** Gives the list of edges which left and right side verify the given conditions.
     An edge is a 3-uple (node1,receptname1), cable , (node2,receptname2) *)
 method edgesSuchThat (lpred: node->bool) (cpred:cable->bool) (rpred: node->bool) : edge list =
   let cable_info c =
       begin
       let (l,r) = (c#get_left,c#get_right) in
       let a = ((self#getNodeByName l.nodename), l.receptname) in
       let b = ((self#getNodeByName r.nodename), r.receptname) in (a,c,b)
       end in
   let l1 = List.map cable_info cables in
   let filter = fun ((n1,_),c,(n2,_)) -> (lpred n1) && (cpred c) && (rpred n2) in
   let l2 = (List.filter filter l1) in l2

 (** Gives the list of edges of a given kind. *)
 method edgesOfKind (lnkind: nodekind option) (ckind: cablekind option) (rnkind: nodekind option) =
   let npred nkind = fun n -> match nkind with None -> true | Some nk -> (n#nodekind  = nk) in
   let cpred kind  = fun c -> match kind  with None -> true | Some ck -> (c#cablekind = ck) in
   self#edgesSuchThat (npred lnkind) (cpred ckind) (npred rnkind)


 (** Gives the list of all edges in the network. *)
 method edges : edge list = self#edgesSuchThat (fun ln->true) (fun c->true) (fun rn->true)

(* ex colore dei cavi incrociati "#52504c" *)

 (** Network translation into the dot language *)
 method dotTrad () =
 let opt = self#dotoptions in
 let labeldistance = opt#labeldistance#get in
 begin
"digraph plan {

"^opt#ratio^"
"^opt#rankdir_for_dot^"
"^opt#nodesep_for_dot^";"^"

/* ***************
        NODES
   *************** */

"^
(StringExtra.Text.to_string
   (List.map
     (fun (n:node)->n#dotTrad opt#iconsize_for_dot)
     (ListExtra.permute opt#shuffler_as_function self#nodes)
   ))
^"
/* ***********************
      DIRECT CABLE EDGES
   *********************** */


edge [dir=none,color=\""^Color.direct_cable^"\",fontsize=8,labelfontsize=8,minlen=1.6,"^
opt#labeldistance_for_dot^",tailclip=true];

"^
(StringExtra.Text.to_string
   (List.map (dotEdgeTrad labeldistance) (self#edgesOfKind None (Some Direct) None)))

^"
/* *********************************
      CROSSOVER/SERIAL CABLE EDGES
   ********************************* */


edge [headclip=true,minlen=1.6,color=\""^Color.crossover_cable^"\",weight=1];

"^
(StringExtra.Text.to_string
   (List.map (dotEdgeTrad labeldistance) (self#edgesOfKind None (Some Crossover) None)))

^
(StringExtra.Text.to_string
   (List.map
      (dotEdgeTrad ~edgeoptions:"style=bold" labeldistance)
      (self#edgesOfKind None (Some NullModem) None)
   ))

^"} //END of digraph\n"

 end (* method dotTrad *)

initializer

 self#set_dotoptions (new Dotoptions.network self);

end


(** {2 Simplified constructors } *)

(** Simplified machine constructor *)
let newMachine ~network name = new machine ~name ~ter:(MSys.string_of_x_policy MSys.HostXServer) ();;

(** Simplified hub constructor (by default 8 ports). *)
let newHub ~network name = new device ~network ~name ~devkind:Hub 8 ();;

(** Simplified switch constructor (by default 8 ports). *)
let newSwitch ~network name = new device ~network ~name ~devkind:Switch 8 ();;

(** Simplified router constructor (by default 8 ports). *)
let newRouter ~network name = new device ~network ~name ~devkind:Router 8 ();;

(** Simplified direct cable constructor.

    {[ Example: newDirectCable "dc1" ("rome","eth0") ("A","port0") ]}   *)
let newDirectCable ~network name ((m1,r1):(nodename*receptname)) ((m2,r2):(nodename*receptname)) =
  new cable ~network ~name ~cablekind:Direct ~left:{nodename=m1;receptname=r1} ~right:{nodename=m2;receptname=r2} ()

(** Simplified crossover cable constructor.

    {[ Example: newCrossoverCable "cc1" ("rome","eth0") ("paris","eth1") ]}  *)
let newCrossoverCable ~network name ((m1,r1):(nodename*receptname)) ((m2,r2):(nodename*receptname)) =
  new cable ~network ~name ~cablekind:Crossover ~left:{nodename=m1;receptname=r1} ~right:{nodename=m2;receptname=r2} ()

(** Simplified cloud with defect parameters. *)
let newCloud ~network name = new cloud ~network ~name ;;


(** {2 Saving and loading a Netmodel.network } *)

(** Pseudo XML now! (using xforest instead of ocamlduce) *)
module Xml = struct

 let network_marshaller = new Oomarshal.marshaller ;;

(** Parse the file containing an xforest representation of the network.
    The given network is updated during the parsing. *)
 let load_network (net:network) (fname:string) =
  let (forest:Xforest.t) = network_marshaller#from_file fname in
  Xforest.print_forest forest;
  match forest with
  | Forest.NonEmpty  (("network", attrs) , childs , Forest.Empty) ->
      net#from_forest ("network", attrs) childs
 ;;

(** Save the xforest representation of the network. *)
let save_network (net:network) (fname:string) =
 Log.print_string "Netmodel.Xml.save_network: begin\n";
 Xforest.print_forest net#to_forest;
 network_marshaller#to_file net#to_forest fname;
 Log.print_string "Netmodel.Xml.save_network: end (success)\n";;

end;; (* module Netmodel.Xml *)


end;; (* module Netmodel *)
